from nipype import Workflow, MapNode, Node
from nipype.interfaces.io import DataSink

import panpipelines.nodes.antstransform as antstransform
import panpipelines.nodes.atlascreate as atlascreate
import panpipelines.nodes.roi_extract as roiextract
from panpipelines.utils.util_functions import *
from panpipelines.utils.transformer import *
import glob

def create(name, wf_base_dir,labels_dict,createGraph=True,execution={},LOGGER=None):
    # Create workflow
    pan_workflow = Workflow(name=name, base_dir=wf_base_dir)

    if LOGGER:
        LOGGER.info(f"Created Workflow {name} with base directory {wf_base_dir}")

    if len(execution.keys()) > 0:
        pan_workflow.config = process_dict(pan_workflow.config,execution)

    # do we have  a mask to exclude?
    mask_list=[]
    maskcreate_node = None
    mask_templates = getParams(labels_dict,"MASK_TEMPLATE")
    if mask_templates:
        mask_name = getParams(labels_dict,"MASK_NAME")
        if not mask_name:
            mask_name="mask"
        mask_index = getParams(labels_dict,"MASK_INDEX")
        if isinstance(mask_templates,list):
            for mask_template in mask_templates:
                evaluated_mask_template = substitute_labels(mask_template,labels_dict)
                if "*" not in evaluated_mask_template:
                    mask_list.append(evaluated_mask_template)
                else:
                    mask_list.extend(glob.glob(evaluated_mask_template))
        else:
            mask_list.extend(glob.glob(mask_templates))

        # if mask template is invalid then continue, otherwise print message and continue without
        # using mask segmenntation
        MASK_TEMPLATE_EXISTS=True
        for mask in mask_list:
            if not os.path.exists(mask):
                MASK_TEMPLATE_EXISTS = False

        if mask_list and MASK_TEMPLATE_EXISTS:    
            # store and restore parameters used by both mask and newatlas
            newatlas_transform_mat = getParams(labels_dict,"NEWATLAS_TRANSFORM_MAT")
            newatlas_transform_ref = getParams(labels_dict,"NEWATLAS_TRANSFORM_REF")
            newatlas_type = getParams(labels_dict,"NEWATLAS_TYPE")
            newatlas_name = getParams(labels_dict,"NEWATLAS_NAME")
            newatlas_probthresh = getParams(labels_dict,"NEWATLAS_PROBTHRESH")
            newatlas_invertroi = getParams(labels_dict,"NEWATLAS_INVERTROI")
            newatlas_indexmode = getParams(labels_dict,"NEWATLAS_INDEXMODE")

            if not getParams(labels_dict,"MASK_TRANSFORM_MAT"):
                labels_dict = updateParams(labels_dict,"MASK_TRANSFORM_MAT",[["identity" for x in mask_list]])

            mask_transform_mat = getParams(labels_dict,"MASK_TRANSFORM_MAT")
            mask_transform_ref = getParams(labels_dict,"MASK_TRANSFORM_REF")
            mask_type = getParams(labels_dict,"MASK_TYPE")
            if not mask_type:
                mask_type = "3D_mask"
            mask_name = getParams(labels_dict,"MASK_NAME")
            mask_probthresh = getParams(labels_dict,"MASK_PROBTHRESH")
            mask_invertroi = getParams(labels_dict,"MASK_INVERTROI")
            mask_indexmode = getParams(labels_dict,"MASK_INDEXMODE")

            labels_dict = updateParams(labels_dict,"NEWATLAS_TRANSFORM_MAT",mask_transform_mat)
            labels_dict = updateParams(labels_dict,"NEWATLAS_TRANSFORM_REF",mask_transform_ref)
            labels_dict = updateParams(labels_dict,"NEWATLAS_TYPE",mask_type)
            labels_dict = updateParams(labels_dict,"NEWATLAS_NAME",mask_name)
            labels_dict = updateParams(labels_dict,"NEWATLAS_PROBTHRESH",mask_probthresh)
            labels_dict = updateParams(labels_dict,"NEWATLAS_INVERTROI",mask_invertroi)
            labels_dict = updateParams(labels_dict,"NEWATLAS_INDEXMODE",mask_indexmode)

            labels_dict = updateParams(labels_dict,"COST_FUNCTION","NearestNeighbor")
            maskcreate_node = atlascreate.create(labels_dict,name=f"maskcreate_{mask_name}_node",roi_list=mask_list,roilabels_list=mask_index,LOGGER=LOGGER)

            labels_dict = updateParams(labels_dict,"NEWATLAS_TRANSFORM_MAT",newatlas_transform_mat,remove_if_none=True)
            labels_dict = updateParams(labels_dict,"NEWATLAS_TRANSFORM_REF",newatlas_transform_ref,remove_if_none=True)
            labels_dict = updateParams(labels_dict,"NEWATLAS_TYPE",newatlas_type,remove_if_none=True)
            labels_dict = updateParams(labels_dict,"NEWATLAS_NAME",newatlas_name,remove_if_none=True)
            labels_dict = updateParams(labels_dict,"NEWATLAS_PROBTHRESH",newatlas_probthresh,remove_if_none=True)
            labels_dict = updateParams(labels_dict,"NEWATLAS_INVERTROI",newatlas_invertroi,remove_if_none=True)
            labels_dict = updateParams(labels_dict,"NEWATLAS_INDEXMODE",newatlas_indexmode,remove_if_none=True)
        else:
            if LOGGER:
                LOGGER.info(f"mask Template defined but valid template file not found. Ignoring mask Template.")

    # do we need to create a custom atlas?
    atlas_index = getParams(labels_dict,"ATLAS_INDEX")
    atlas_file = getParams(labels_dict,"ATLAS_FILE")
    atlas_name = getParams(labels_dict,"ATLAS_NAME")
    atlascreate_node=None
    if not atlas_file:
        newatlas_list=[]
        newatlas_templates = getParams(labels_dict,"NEWATLAS_TEMPLATE")
        newatlas_index = getParams(labels_dict,"NEWATLAS_INDEX")
        if not atlas_name:
            atlas_name = getParams(labels_dict,"NEWATLAS_NAME")
        if isinstance(newatlas_templates,list):
            for newatlas_template in newatlas_templates:
                evaluated_newatlas_template = substitute_labels(newatlas_template,labels_dict)
                if "*" not in evaluated_newatlas_template:
                    newatlas_list.append(evaluated_newatlas_template)
                else:
                    newatlas_list.extend(glob.glob(evaluated_newatlas_template))
        else:
            newatlas_list.extend(glob.glob(newatlas_templates))

        NEWATLAS_TEMPLATE_SORT = isTrue(getParams(labels_dict,"NEWATLAS_TEMPLATE_SORT"))
        if NEWATLAS_TEMPLATE_SORT:
            newatlas_list.sort()
            
        labels_dict = updateParams(labels_dict,"COST_FUNCTION","NearestNeighbor")
        atlascreate_node = atlascreate.create(labels_dict,name=f"atlascreate_{atlas_name}_node",roi_list=newatlas_list,roilabels_list=newatlas_index,LOGGER=LOGGER)

    roimean_node = roiextract.create(labels_dict,name="subject_metrics",LOGGER=LOGGER)
    roimean_map_node = MapNode(roimean_node.interface,name="subject_metrics_map",iterfield=['input_file'])

    if atlascreate_node:
        pan_workflow.connect(atlascreate_node,'atlas_index',roimean_map_node,'atlas_index')
    else:
        roimean_map_node.inputs.atlas_index = atlas_index


    atlas_transform_mat=getParams(labels_dict,"ATLAS_TRANSFORM_MAT") 
    atlas_transform_ref=getParams(labels_dict,"ATLAS_TRANSFORM_REF")   
    # should we transform atlas?
    if atlas_transform_mat is not None:
        labels_dict = updateParams(labels_dict,"COST_FUNCTION","NearestNeighbor")
        if atlascreate_node:
            atlas_transform_node = antstransform.create(labels_dict,name="atlas_transform", trans_mat=atlas_transform_mat,ref_file=atlas_transform_ref, LOGGER=LOGGER)
            pan_workflow.connect(atlascreate_node,'atlas_file',atlas_transform_node,'input_file')    
        else:
            atlas_transform_node = antstransform.create(labels_dict,name="atlas_transform",input_file=atlas_file, trans_mat=atlas_transform_mat,ref_file=atlas_transform_ref, LOGGER=LOGGER)
        pan_workflow.connect(atlas_transform_node,'out_file',roimean_map_node,'atlas_file')
    else:
        if atlascreate_node:
            pan_workflow.connect(atlascreate_node,'atlas_file',roimean_map_node,'atlas_file') 
        else:
            roimean_map_node.inputs.atlas_file = atlas_file

    measures_list=[]
    measures_template = getParams(labels_dict,"MEASURES_TEMPLATE")
    if isinstance(measures_template,list):
        for meas_template in measures_template:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict)
            measures_list.extend(glob.glob(evaluated_meas_template))
    else:
        evaluated_measures_template = substitute_labels(measures_template,labels_dict)
        measures_list.extend(glob.glob(evaluated_measures_template))
        
    measures_list.sort()

    # should we transform measures?
    measures_transform_mat=getParams(labels_dict,"MEASURES_TRANSFORM_MAT")
    measures_transform_ref=getParams(labels_dict,"MEASURES_TRANSFORM_REF")
    if measures_transform_mat is not None:
        labels_dict = removeParam(labels_dict,"COST_FUNCTION")
        labels_dict = removeParam(labels_dict,"OUTPUT_TYPE")
        measures_transform_node = antstransform.create(labels_dict,name="measure_transform",trans_mat=measures_transform_mat,ref_file=measures_transform_ref,LOGGER=LOGGER)
        measures_transform_map_node = MapNode(measures_transform_node.interface,name="measure_transform_map",iterfield=['input_file'])
        measures_transform_map_node.inputs.input_file = measures_list
        pan_workflow.connect(measures_transform_map_node,'out_file',roimean_map_node,'input_file')          
    else:
        roimean_map_node.inputs.input_file = measures_list

    if maskcreate_node:
        pan_workflow.connect(maskcreate_node,'atlas_file',roimean_map_node,'mask_file')
    elif mask_list and MASK_TEMPLATE_EXISTS:
        if len(mask_list) == 1 and not getParams(labels_dict,"MASK_TRANSFORM_MAT"):
            roimean_map_node.inputs.mask_file = mask_list[0]
        else:
            LOGGER.warn(f"Something went wrong. Should not have mask list: {mask_list} with transform at this point")


    sinker_dir = getParams(labels_dict,"SINKDIR")
    if sinker_dir:
        sinker = Node(DataSink(),name='roiextract_sink')
        sinker_basedir = os.path.dirname(sinker_dir)
        sinker_folder = os.path.basename(sinker_dir)
        if not os.path.exists(sinker_basedir):
            os.makedirs(sinker_basedir,exist_ok=True)
        sinker.inputs.base_directory = sinker_basedir

        measure_count=0
        substitutions=[]
        for measure_name in measures_list:
            measure_name_parts = os.path.basename(measure_name).split("_")
            if len(measure_name_parts) > 1:
                measure_name_stub = "_".join(measure_name_parts[-2:])
            else:
                measure_name_stub = measure_name_parts[-1]

            measure_name_suffix = measure_name_stub.split(".")[0]
            substitutions+=[("_subject_metrics_map" +str(measure_count),measure_name_suffix)]
            measure_count = measure_count + 1 
        if substitutions:
            sinker.inputs.substitutions = substitutions

        pan_workflow.connect( roimean_map_node,"roi_csv",sinker,f"{sinker_folder}")
        pan_workflow.connect( roimean_map_node,"roi_csv_metadata",sinker,f"{sinker_folder}.@metadata")
        pan_workflow.connect( roimean_map_node,"mask_file",sinker,f"{sinker_folder}.@maskfile")
        pan_workflow.connect( roimean_map_node,"html_file",sinker,f"{sinker_folder}.@htmlfile")
        pan_workflow.connect( roimean_map_node,"roi_csv_sizes",sinker,f"{sinker_folder}.@roi_csv_sizes")
        pan_workflow.connect( roimean_map_node,"roi_csv_coverage",sinker,f"{sinker_folder}.@roi_csv_coverage")

    if createGraph:
         pan_workflow.write_graph(graph2use='flat')

    return pan_workflow



from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def dummy_proc(labels_dict):
    cwd=os.getcwd()
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    participant_session = getParams(labels_dict,'PARTICIPANT_SESSION')
    dummy_file = os.path.join(cwd,f'sub-{participant_label}_ses-{participant_session}_log.txt')
    
    with open(dummy_file,"w") as outfile:
        outfile.write(f"sub-{participant_label}_ses-{participant_session} finished")
        
    out_files=[]
    out_files.append(dummy_file)

    return {
        "out_files" : out_files
    }

class dummyInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)

class dummyOutputSpec(TraitedSpec):
    out_files = traits.List(desc='list of files')
    
class dummy_pan(BaseInterface):
    input_spec = dummyInputSpec
    output_spec = dummyOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = dummy_proc(
            self.inputs.labels_dict,
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='dummy_node',LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(dummy_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
   
    return pan_node





from panpipelines.utils.util_functions import *
from panpipelines.utils.transformer import *
from nipype import logging as nlogging
import pysftp

UTLOGGER=nlogging.getLogger('nipype.utils')

def ftp_exists(remote_path,hostname,username,password,port):
    with pysftp.Connection(host=hostname,username=username,password=password,port=port) as connection:
        return connection.exists(remote_path)
  
def ftp_listdir(remote_path,hostname,username,password,port):
    files_and_dirs=[]
    with pysftp.Connection(host=hostname,username=username,password=password,port=port) as connection:
        files_and_dirs = connection.listdir(remote_path)
    return files_and_dirs

def ftp_listdir_attr(remote_path,hostname,username,password,port):
    files_and_dirs=[]
    with pysftp.Connection(host=hostname,username=username,password=password,port=port) as connection:
        files_and_dirs  =  connection.listdir_attr(remote_path)
    return files_and_dirs

def ftp_download(remote_path, target_local_path,hostname,username,password,port):
    try:
        # Create the target directory if it does not exist
        if not os.path.isdir(os.path.dirname(target_local_path)):
            try:
                os.makedirs(os.path.dirname(target_local_path),exist_ok=True)
            except Exception as err:
                raise Exception(err)
        with pysftp.Connection(host=hostname,username=username,password=password,port=port) as connection:
            connection.get(remote_path, target_local_path)
    except Exception as err:
        raise Exception(err)

def ftp_downloaddir_recursive(remote_path, target_local_path,hostname,username,password,port):
    try:
        # Create the target directory if it does not exist
        path, _ = os.path.split(target_local_path)
        if not os.path.isdir(path):
            try:
                os.makedirs(path,exist_ok=True)
            except Exception as err:
                raise Exception(err)
        with pysftp.Connection(host=hostname,username=username,password=password,port=port) as connection:
            connection.chdir(os.path.dirname(remote_path))
            basefolder = os.path.basename(remote_path)
            connection.get_r(basefolder, path)
    except Exception as err:
        raise Exception(err)

def ftp_upload(source_local_path, remote_path,hostname,username,password,port):
    try:
        with pysftp.Connection(host=hostname,username=username,password=password,port=port) as connection:
            if not connection.exists(os.path.dirname(remote_path)):
                connection.makedirs(os.path.dirname(remote_path))
            connection.put(source_local_path, remote_path)
    except Exception as err:
        raise Exception(err)

def ftp_uploaddir_recursive(source_local_path, remote_path,hostname=None,username=None,password=None,port=None):
    try:
        with pysftp.Connection(host=hostname,username=username,password=password,port=port) as connection:
            if not connection.exists(remote_path):
                connection.makedirs(remote_path)
            connection.put_r(source_local_path, remote_path)
    except Exception as err:
        raise Exception(err)


def ftp_upload_subjectbids(subject_dir, remote_path,hostname=None,username=None,password=None, port=None, replace=True):
    subject = os.path.basename(subject_dir)
    print(f"uploading {subject} from {subject_dir} to {remote_path}.")
    if ftp_exists(remote_path,hostname,username,password,port) and replace:
        print(f"{remote_path} already exists. Will remove before uploading {subject}")
        ftp_deletedir_recursive(remote_path,hostname,username,password,port)
        print(f"{remote_path} successfully deleted.")
        
    ftp_uploaddir_recursive(subject_dir, remote_path,hostname=hostname,username=username,password=password,port=port)
    print(f"{subject} successfully uploaded.")

def ftp_upload_allbids(bids_dir, remote_path,hostname=None,username=None,password=None,port=None):
    ftp_uploaddir_recursive(bids_dir,remote_path,hostname,username,password,port)
    remote_participantsTSV=os.path.join(remote_path,"participants.tsv")
    local_participantsTSV = os.path.join(bids_dir,"participants.tsv")
    metadata = {}
    local_metadata_file = newfile(assocfile=local_participantsTSV,extension="json")
    if os.path.exists(local_metadata_file):
        with open(local_metadata_file,"r") as infile:
            metadata = json.load(infile)

    history={}
    history["SourceBIDS"]=f"{bids_dir}"
    history["TargetBIDS"]=f"{remote_path}"
    date_uploaded = get_datetimestring_utc()
    history["Description"]=f"All bids files uploaded from {bids_dir}"
    r_file = newfile(assocfile=remote_participantsTSV,extension="json")
    tmpJsonFile = get_upload_metadata(r_file, local_participantsTSV,remote_participantsTSV,date_uploaded,metadata=metadata,history=history)
    ftp_upload(tmpJsonFile,r_file,hostname,username,password,port)

def ftp_delete(remote_path,hostname,username,password,port):
    try:
        with pysftp.Connection(host=hostname,username=username,password=password,port=port) as connection:
            connection.remove(remote_path)
    except Exception as err:
        raise Exception(err)

def ftp_deletedir(remote_path,hostname,username,password,port):
    try:
        with pysftp.Connection(host=hostname,username=username,password=password,port=port) as connection:
            connection.rmdir(remote_path)
    except Exception as err:
        raise Exception(err)

def ftp_deletedir_recursive(remote_path,hostname,username,password,port):
    dirs=[remote_path]
    try:
        with pysftp.Connection(host=hostname,username=username,password=password,port=port) as connection:
            connection.walktree(dirs[0], fcallback=connection.remove, dcallback=dirs.append, ucallback=connection.remove, recurse=True)
            for d in reversed(dirs):
                connection.rmdir(d)
    except Exception as err:
        raise Exception(err)


def get_upload_metadata(metadata_file, upload_source = None, upload_destination=None, date_uploaded = None, creator=None, metadata={},history={}):

    metadata = updateParams(metadata,"MetadataFile",f"{metadata_file}")

    if not date_uploaded:
        date_uploaded = get_datetimestring_utc()
    metadata["LastUpload"] = date_uploaded

    if not "History" in metadata.keys():
        metadata["History"]={}

    if not date_uploaded in metadata["History"].keys():
        metadata["History"][date_uploaded] = {}

    if upload_source:
        metadata = updateParams(metadata,"UploadSource",f"{upload_source}")
        metadata["History"][date_uploaded]["UploadSource"] = f"{upload_source}"

    if upload_destination:
        metadata = updateParams(metadata,"UploadDestination",f"{upload_destination}")
        metadata["History"][date_uploaded]["UploadDestination"] = f"{upload_destination}"


    if not creator:
        if "USER" in os.environ.keys():
            metadata["History"][date_uploaded]["UploadedBy"] = os.environ["USER"]
        else:
            metadata["History"][date_uploaded]["UploadedBy"] = "Unknown"

    if "HOSTNAME" in os.environ.keys():
        metadata["History"][date_uploaded]["Hostname"] = os.environ["HOSTNAME"]

    metadata["History"][date_uploaded]["IP"] = get_ip()

    for itemkey, itemvalue in history.items():
        metadata["History"][date_uploaded][itemkey]=itemvalue
        
    tmp_json = tempfile.mkstemp()[1] + os.path.basename(metadata_file)
    export_labels(metadata,tmp_json)

    return tmp_json

def upload_metadata(local_metadata_file,remote_metadata_file,upload_source,upload_destination,metadata={},history={},date_uploaded=None,hostname=None,username=None,password=None, port=None):
    try:
        tmp_json_file = tempfile.mkstemp()[1] + os.path.basename(remote_metadata_file)
        ftp_download(remote_metadata_file,tmp_json_file,hostname,username,password,port)
        if tmp_json_file and os.path.exists(tmp_json_file):
            with open(tmp_json_file,'r') as infile:
                prev_metadata = json.load(infile)
                for itemkey,itemvalue in prev_metadata.items():
                    if not itemkey in metadata.keys():
                        metadata[itemkey]=itemvalue
    except Exception as e:
        pass
        
    tmpJsonFile = get_upload_metadata(remote_metadata_file,upload_source ,upload_destination,date_uploaded,metadata=metadata,history=history)
    ftp_upload(tmpJsonFile,remote_metadata_file,hostname,username,password,port)

    shutil.copy(tmpJsonFile,local_metadata_file)

    return local_metadata_file



__all__ = [
    "panscript",
    "fmriprep_panscript",
    "aslprep_panscript",
    "pancontainer_panscript",
    "sdcflows_fieldmap",
    "mne_make_surfaces",
    "paninfo_bidsqc",
    "additional_transforms",
    "ftp_upload_pan",
    "pandb_download"
]



from nipype import Workflow, MapNode, Node
from nipype.interfaces.io import DataSink

import panpipelines.nodes.antstransform as antstransform
import panpipelines.nodes.parse_textdata as parse_textdata
from panpipelines.utils.util_functions import *
from panpipelines.utils.transformer import *
import glob

def create(name, wf_base_dir,labels_dict,createGraph=True,execution={},LOGGER=None):
    # Create workflow
    pan_workflow = Workflow(name=name, base_dir=wf_base_dir)

    if LOGGER:
        LOGGER.info(f"Created Workflow {name} with base directory {wf_base_dir}")

    if len(execution.keys()) > 0:
        pan_workflow.config = process_dict(pan_workflow.config,execution)


    parsetextdata_node = parse_textdata.create(labels_dict,name="subject_text",LOGGER=LOGGER)
    parsetextdata_map_node = MapNode(parsetextdata_node.interface,name="subject_text_map",iterfield=['textdata','textdata_type','custom_prefix'])

    filecount_list=[]
    measures_list=[]
    measures_template = getParams(labels_dict,"MEASURES_TEMPLATE")
    if isinstance(measures_template,list):
        for meas_template in measures_template:
            evaluated_meas_template = substitute_labels(meas_template,labels_dict)
            meas_files=glob.glob(evaluated_meas_template)
            measures_list.extend(meas_files)
            filecount_list.append(len(meas_files))
    else:
        measures_list.extend(glob.glob(measures_template))
    measures_len = len(measures_list)

    measures_texttype_list=["" for x in range(measures_len)]
    measures_texttypes = getParams(labels_dict,"MEASURES_TEXTTYPES")
    file_index=0
    for meas_file in measures_list:
        if measures_texttypes and isinstance(measures_texttypes,dict):
            for itemkey, itemvalue in measures_texttypes.items():
                if substitute_labels(itemkey,labels_dict) in os.path.basename(meas_file):
                    measures_texttype_list[file_index] =substitute_labels(itemvalue,labels_dict)
                    break
        file_index=file_index + 1


    measures_prefix_list=["" for x in range(measures_len)]
    measures_prefixes = getParams(labels_dict,"MEASURES_PREFIXES")
    file_index=0
    for meas_file in measures_list:
        if measures_prefixes and isinstance(measures_prefixes,dict):
            for itemkey, itemvalue in measures_prefixes.items():
                if substitute_labels(itemkey,labels_dict) in os.path.basename(meas_file):
                    measures_prefix_list[file_index] = substitute_labels(itemvalue,labels_dict)
                    break
        file_index=file_index + 1

    parsetextdata_map_node.inputs.textdata = measures_list
    parsetextdata_map_node.inputs.textdata_type = measures_texttype_list
    parsetextdata_map_node.inputs.custom_prefix = measures_prefix_list
    pan_workflow.add_nodes([parsetextdata_map_node])

    sinker_dir = getParams(labels_dict,"SINKDIR")
    if sinker_dir:
        sinker = Node(DataSink(),name='textmeasures_sink')
        sinker_basedir = os.path.dirname(sinker_dir)
        sinker_folder = os.path.basename(sinker_dir)
        if not os.path.exists(sinker_basedir):
            os.makedirs(sinker_basedir,exist_ok=True)
        sinker.inputs.base_directory = sinker_basedir

        measure_count=0
        substitutions=[]
        for measure_name in measures_list:
            measure_name_parts = os.path.basename(measure_name).split("_")
            if len(measure_name_parts) > 1:
                measure_name_stub = measure_name_parts[-2]
            else:
                measure_name_stub = measure_name_parts[-1]

            measure_name_suffix = measure_name_stub.split(".")[0]
            substitutions+=[("_subject_text_map" + str(measure_count),measure_name_suffix)]
            measure_count = measure_count + 1 
        if substitutions:
            sinker.inputs.substitutions = substitutions

        pan_workflow.connect( parsetextdata_map_node,"roi_csv",sinker,f"{sinker_folder}")
        pan_workflow.connect( parsetextdata_map_node,"roi_csv_metadata",sinker,f"{sinker_folder}.@metadata")


    if createGraph:
         pan_workflow.write_graph(graph2use='flat')


    return pan_workflow



from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import panpipelines.utils.transformer as tr
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging

IFLOGGER=nlogging.getLogger('nipype.interface')

def tractseg_proc(labels_dict,input_dir):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)

    ALIGN_APPROACH=getParams(labels_dict,"TRACTSEG_ALIGN_APPROACH")
    if not ALIGN_APPROACH:
        ALIGN_APPROACH="FA"

    TRACTSEG_HOME=getParams(labels_dict,"TRACTSEG_HOME")
    if not TRACTSEG_HOME:
        TRACTSEG_HOME=os.path.join(cwd,"tractseg_home")
        os.makedirs(TRACTSEG_HOME,exist_ok=True)  
        labels_dict = updateParams(labels_dict,"TRACTSEG_HOME",TRACTSEG_HOME) 

    # set up initial vars

    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    participant_session = getParams(labels_dict,'PARTICIPANT_SESSION')

    input_dir =substitute_labels(input_dir,labels_dict)
    output_dir =substitute_labels("<CWD>/tractseg_out", labels_dict)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir,exist_ok=True)

    work_dir = substitute_labels("<CWD>/tractseg_work", labels_dict)
    if not os.path.exists(work_dir):
        os.makedirs(work_dir,exist_ok=True)

    prep_dir = substitute_labels("<CWD>/tractseg_prep", labels_dict)
    if not os.path.exists(prep_dir):
        os.makedirs(prep_dir,exist_ok=True)

    # mrtrix container - check version
    mr_command_base,mr_container = getContainer(labels_dict,nodename="tensor", SPECIFIC="MRTRIX_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the mrtrix version:")
    command = f"{mr_command_base} mrconvert --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    if mr_container:
        IFLOGGER.info("\nChecking the container version:")
        command = f"{mr_command_base} --version"
        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)
    
    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["APPTAINER_TEMPLATEFLOW_HOME"]=translate_binding(mr_command_base,TEMPLATEFLOW_HOME)

    # get qsiprep outputs and transforms - prepare data for TractSeg
    resolution="1"
    struct_mni2009 = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'anat','*_from-T1w_to-MNI152NLin2009cAsym_mode-image_xfm.h5'))
    mni2009_mni = tr.get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin6Asym",suffix="xfm",extension=[".h5"])

    fslmni=os.path.join(f"/opt/fsl/data/standard/MNI152_T1_{resolution}mm.nii.gz")
    fslFAmni=os.path.join(f"/opt/fsl/data/standard/FMRIB58_FA_{resolution}mm.nii.gz")
    mni152=get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin6Asym",resolution=int(resolution),suffix="T1w",extension=[".nii.gz"])
    mni2009=get_template_ref(TEMPLATEFLOW_HOME,"MNI152NLin2009cAsym",resolution=int(resolution),suffix="T1w",extension=[".nii.gz"])
    t1acpc = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'anat','*_desc-preproc_T1w.nii.gz'))
    mask = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(participant_session),'dwi','*_desc-brain_mask.nii.gz'))
    dwiacpc = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(participant_session),'dwi','*_desc-preproc_dwi.nii.gz'))
    dwigrad = getGlob(os.path.join(input_dir,'sub-{}'.format(participant_label),'ses-{}'.format(participant_session),'dwi','*_desc-preproc_dwi.b'))
    dwitrix=newfile(outputdir=prep_dir,assocfile=dwiacpc,extension=".mif")

    IFLOGGER.info("Convert qsiprep dwi to mrtrix format.")
    command=f"{mr_command_base} mrconvert {dwiacpc} {dwitrix} -grad {dwigrad}"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    if ALIGN_APPROACH == "FA":
        IFLOGGER.info("Create Affine transforms from FA.")

        # convert dwitrix to LAS - flirt really needs data in standard orientation to work
        dwitrix_las=newfile(assocfile=dwitrix,intwix="ori-LAS")
        command=f"{mr_command_base} mrconvert {dwitrix} {dwitrix_las} -strides -1,+2,+3,+4"
        runCommand(command)

        tensor_mrtrix = os.path.join(prep_dir,'sub-{}_tensor.mif'.format(participant_label))
        command=f"{mr_command_base} dwi2tensor -mask {mask} {dwitrix_las} {tensor_mrtrix}"
        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

        fa_mrtrix = newfile(assocfile=dwitrix_las,suffix="fa")
        command=f"{mr_command_base} tensor2metric -mask {mask} -fa {fa_mrtrix} {tensor_mrtrix}"
        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

        fa_fsl = newfile(assocfile=fa_mrtrix,extension=".nii.gz")
        command=f"{mr_command_base} mrconvert {fa_mrtrix} {fa_fsl}"
        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

        fa_fsl_mni=newfile(outputdir=prep_dir,assocfile=fa_fsl,intwix="space-MNI152NLin6Asym")
        trans_affine_fsl=os.path.join(prep_dir,f"{participant_label}_dwiacpc_MNI152_fsl.mat")
        tr.fsl_reg_flirt(fa_fsl,fslFAmni,fa_fsl_mni,trans_affine_fsl,mr_command_base,dof="6",cost="mutualinfo")

        inverse_trans_affine_fsl=newfile(outputdir=prep_dir,assocfile=trans_affine_fsl,suffix="inverse")
        tr.invertAffine_FLIRT(trans_affine_fsl, inverse_trans_affine_fsl, mr_command_base)

        trans_affine=newfile(outputdir=prep_dir,assocfile=trans_affine_fsl,suffix="mrtrix")
        command=f"{mr_command_base} transformconvert {trans_affine_fsl} {dwitrix_las} {fslFAmni} flirt_import {trans_affine}"
        runCommand(command)

        dwimni152_las=newfile(outputdir=prep_dir,assocfile=dwitrix_las,intwix="space-MNI152NLin6Asym")
        command=f"{mr_command_base} mrtransform -template {fslFAmni} -strides {fslFAmni} -linear {trans_affine}  {dwitrix_las} {dwimni152_las}"
        runCommand(command)

        dwimni152_las_nii=newfile(assocfile=dwimni152_las,extension=".nii.gz")
        command=f"{mr_command_base} mrconvert {dwimni152_las} {dwimni152_las_nii}"
        runCommand(command)

        gradmni152=newfile(assocfile=dwimni152_las_nii,extension=".b")
        bval=newfile(assocfile=dwimni152_las_nii,extension=".bval")
        bvec=newfile(assocfile=dwimni152_las_nii,extension=".bvec")
        command=f"{mr_command_base} dwigradcheck {dwimni152_las} -export_grad_mrtrix {gradmni152}"
        runCommand(command)
        command=f"{mr_command_base} dwigradcheck {dwimni152_las} -export_grad_fsl {bvec} {bval}"
        runCommand(command)

    else:
        IFLOGGER.info("Create Affine transforms from existing qsiprep transforms.")
        os.chdir(prep_dir)
        t1 = tr.disassembleTransforms(struct_mni2009,"t1",mr_command_base)
        t1mat=[x for x in t1 if x.endswith(".mat")][0]
        IFLOGGER.info(f"Created Affine transforms {t1mat}.")

        t2=tr.disassembleTransforms(mni2009_mni,"t2",mr_command_base)
        t2mat=[x for x in t2 if x.endswith(".mat")][0] 
        IFLOGGER.info(f"Created Affine transforms {t2mat}.")
        os.chdir(cwd)

        t1_fsl=newfile(outputdir=prep_dir,assocfile=t1mat,suffix="fsl")
        tr.convert_affine_ants_to_fsl(t1mat,dwiacpc,mni2009,t1_fsl)

        t2_fsl=newfile(outputdir=prep_dir,assocfile=t2mat,suffix="fsl")
        tr.convert_affine_ants_to_fsl(t2mat,mni2009,mni152,t2_fsl)

        trans_affine_fsl=os.path.join(prep_dir,f"{participant_label}_dwiacpc_MNI152_fsl.mat")
        tr.concatAffine_FLIRT(t1_fsl,t2_fsl, trans_affine_fsl, mr_command_base)

        inverse_trans_affine_fsl=newfile(outputdir=prep_dir,assocfile=trans_affine_fsl,suffix="inverse")
        tr.invertAffine_FLIRT(trans_affine_fsl, inverse_trans_affine_fsl, mr_command_base)

        trans_affine=newfile(outputdir=prep_dir,assocfile=trans_affine_fsl,suffix="mrtrix")
        command=f"{mr_command_base} transformconvert {trans_affine_fsl} {dwiacpc} {mni152} flirt_import {trans_affine}"
        runCommand(command)

        dwimni152=newfile(outputdir=prep_dir,assocfile=dwitrix,intwix="space-MNI152NLin6Asym")
        command=f"{mr_command_base} mrtransform -force -template {mni152} -linear {trans_affine}  {dwitrix} {dwimni152}"
        runCommand(command)

        dwimni152_las=newfile(assocfile=dwimni152,intwix="ori-LAS")
        command=f"{mr_command_base} mrconvert {dwimni152} {dwimni152_las} -strides -1,+2,+3,+4"
        runCommand(command)

        dwimni152_las_nii=newfile(assocfile=dwimni152_las,extension=".nii.gz")
        command=f"{mr_command_base} mrconvert {dwimni152_las} {dwimni152_las_nii}"
        runCommand(command)

        gradmni152=newfile(assocfile=dwimni152_las_nii,extension=".b")
        bval=newfile(assocfile=dwimni152_las_nii,extension=".bval")
        bvec=newfile(assocfile=dwimni152_las_nii,extension=".bvec")
        command=f"{mr_command_base} dwigradcheck {dwimni152_las} -export_grad_mrtrix {gradmni152}"
        runCommand(command)
        command=f"{mr_command_base} dwigradcheck {dwimni152_las} -export_grad_fsl {bvec} {bval}"
        runCommand(command)

    # Run tractseg
    command_base, container = getContainer(labels_dict,nodename="tractseg", SPECIFIC="TRACTSEG_CONTAINER",LOGGER=IFLOGGER)
    IFLOGGER.info("Checking the tractseg version:")
    command = f"{command_base} TractSeg --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["APPTAINER_TEMPLATEFLOW_HOME"]=translate_binding(command_base,TEMPLATEFLOW_HOME)

    EXTRA_PARAMS = getParams(labels_dict,"EXTRA_TRACTSEG_PARAMS")
    if not EXTRA_PARAMS:
        EXTRA_PARAMS=""

    # generate tensors using mrtrix, use grad from qsiprep
    params="-i "+dwimni152_las_nii+\
        " -o "+output_dir+\
        " --bvals "+bval+\
        " --bvecs "+bvec+\
        " --raw_diffusion_input"

    command=f"{command_base} TractSeg"\
            " "+params + " " + EXTRA_PARAMS

    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    bundles=glob.glob(os.path.join(output_dir,"bundle_segmentations","*.nii.gz"))
    bundles_native_bin_dir=os.path.join(output_dir,"bundles_segmentations_native_bin")
    if not os.path.exists(bundles_native_bin_dir):
        os.makedirs(bundles_native_bin_dir,exist_ok=True)
    
    STORE_PROBS = False
    if "get_probabilities" in command:
        STORE_PROBS = True

    if STORE_PROBS:
        bundles_native_prob_dir=os.path.join(output_dir,"bundles_segmentations_native_prob")
        if not os.path.exists(bundles_native_prob_dir):
            os.makedirs(bundles_native_prob_dir,exist_ok=True)

    TRACTSEG_PROBTHRESH=getParams(labels_dict,"TRACTSEG_PROBTHRESH")
    if not TRACTSEG_PROBTHRESH:
        TRACTSEG_PROBTHRESH="0.5"

    for bundle in bundles:
        if STORE_PROBS:
            bundle_native=newfile(outputdir=bundles_native_prob_dir,assocfile=bundle)
            if ALIGN_APPROACH == "FA":        
                applyAffine_flirt(bundle,fa_fsl,bundle_native,inverse_trans_affine_fsl,mr_command_base,interp="trilinear")
            else:
                applyAffine_flirt(bundle,dwiacpc,bundle_native,inverse_trans_affine_fsl,mr_command_base,interp="trilinear")
            command=f"{mr_command_base} mrconvert -force {bundle_native} {bundle_native} -strides -1,-2,+3"
            runCommand(command)

            bundle_native_bin=newfile(outputdir=bundles_native_bin_dir,assocfile=bundle)
            command=f"{mr_command_base} fslmaths {bundle_native} -thr {TRACTSEG_PROBTHRESH} -bin {bundle_native_bin}"
            runCommand(command)
        else:
            bundle_native_bin=newfile(outputdir=bundles_native_bin_dir,assocfile=bundle)
            if ALIGN_APPROACH == "FA":        
                applyAffine_flirt(bundle,fa_fsl,bundle_native_bin,inverse_trans_affine_fsl,mr_command_base,interp="nearestneighbour")
            else:
                applyAffine_flirt(bundle,dwiacpc,bundle_native_bin,inverse_trans_affine_fsl,mr_command_base,interp="nearestneighbour")

            command=f"{mr_command_base} mrconvert -force {bundle_native_bin} {bundle_native_bin} -strides -1,-2,+3"
            runCommand(command)

    out_files=[]
    out_files.insert(0,inverse_trans_affine_fsl)
    out_files.insert(1,dwiacpc)
    out_files.insert(2,dwimni152_las_nii)

    return {
        "output_dir":cwd,
        "out_files":out_files
    }



class tractsegInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    input_dir = traits.String("",desc="QSIPREP Output Directory", usedefault=True)

class tractsegOutputSpec(TraitedSpec):
    output_dir = traits.String(desc="Tractseg output directory")
    out_files = traits.List(desc='list of files')
    
class tractseg_pan(BaseInterface):
    input_spec = tractsegInputSpec
    output_spec = tractsegOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = tractseg_proc(
            self.inputs.labels_dict,
            self.inputs.input_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name="tractseg_node",input_dir="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(tractseg_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict

    if input_dir is None:
        input_dir = ""
        
    pan_node.inputs.input_dir =  input_dir

    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging
import panpipelines.scripts.pancontainer_panscript as pancontainer_script
from panpipelines import Factory

IFLOGGER=nlogging.getLogger('nipype.interface')


def pancontainergroup_proc(labels_dict):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)

    datelabel = datetime.datetime.now().strftime("%Y%m%d_%H%M%S_%f")
    labels_dict = updateParams(labels_dict,"DATELABEL",datelabel)

    pipeline_config_loc= getParams(labels_dict,"PIPELINE_CONFIG_LOC")
    if not pipeline_config_loc:
        pipeline_config_loc = f"<CWD>/pipeline_config_file_{datelabel}.config"
        labels_dict = updateParams(labels_dict,"PIPELINE_CONFIG_LOC",pipeline_config_loc)

    pipeline_config_loc = substitute_labels(f"{pipeline_config_loc}",labels_dict)
    export_labels(labels_dict,pipeline_config_loc)

    panFactory = Factory.getPANFactory()
    script_class= getParams(labels_dict,"SCRIPT_CLASS")
    script_file= getParams(labels_dict,"SCRIPT_FILE")
    if script_file:       
        new_script_file = newfile(assocfile=script_file,prefix="pantransformed")
        with open(script_file,"r") as infile:
            script_lines = infile.readlines()

        new_script_lines =[]
        for x in script_lines:
            new_script_lines.append(special_substitute_labels(x,labels_dict))

        with open(new_script_file,"w") as outfile:
            outfile.writelines(new_script_lines)

        script_file = new_script_file

    elif script_class:
        run_script = panFactory.get_runscript(script_class)
        script_file = run_script.__file__

    elif LOGGER:
        LOGGER.warn(f"<SCRIPT_CLASS> and <SCRIPT_FILE> not defined. Cannot proceed")
        sys.exit(1)
    else:
        print(f"<SCRIPT_CLASS> and <SCRIPT_FILE> not defined. Cannot proceed")
        sys.exit(1)       


    # Specify script inputs
    params = special_substitute_labels(getParams(labels_dict,"PANCONTAINER_PARAMS"),labels_dict)
    main_command = special_substitute_labels(getParams(labels_dict,"PANCONTAINER_COMMANDS"),labels_dict)
    if not main_command:
        main_command = "python"

    CONTAINER_TO_USE = getParams(labels_dict,"CONTAINER_TO_USE")
    script_interactive= isTrue(getParams(labels_dict,"SCRIPT_INTERACTIVE"))
    try:
        if CONTAINER_TO_USE:
            panscript = pancontainer_script.pancontainer_panscript(labels_dict,params=params,command=f"{main_command} {script_file}",interactive=script_interactive, container_img=CONTAINER_TO_USE)
        else: 
            panscript = pancontainer_script.pancontainer_panscript(labels_dict,params=params,interactive=script_interactive,command=f"{main_command} {script_file}")
        panscript.run()
    except Exception as e:
        IFLOGGER.error(f"Exception thrown by {script_file} script {e}")


    with open(pipeline_config_loc,"r") as infile:
        pipeline_config_json = json.load(infile)

    if pipeline_config_json and "METADATA_FILE" in pipeline_config_json.keys():
        metadata_file = pipeline_config_json["METADATA_FILE"]
    elif "METADATA_FILE" in labels_dict:
        metadata_file = labels_dict["METADATA_FILE"]
    else:
        metadata_file = os.path.join(cwd,f'{os.path.basename(script_file)}_metadata.json')
        with open(metadata_file,"w") as outfile:
            outfile.write(f"No metadata file created by script")

    if pipeline_config_json and "OUTPUT_FILE" in pipeline_config_json.keys():
        output_file = pipeline_config_json["OUTPUT_FILE"]
    elif "OUTPUT_FILE" in labels_dict:
        output_file = labels_dict["OUTPUT_FILE"]
    else:
        output_file = os.path.join(cwd,f'{os.path.basename(script_file)}_output.json')
        with open(output_file,"w") as outfile:
            outfile.write(f"No output file created by script")
    
    out_files=[]
    out_files.insert(0,output_file)
    out_files.insert(1,metadata_file)
    out_files.insert(2,script_file)
    out_files.insert(3,pipeline_config_loc)

    return {
        "output_file" : output_file,
        "metadata_file" : metadata_file,
        "pipeline_config" :pipeline_config_loc,
        "out_files":out_files
    }


class pancontainergroupInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)

class pancontainergroupOutputSpec(TraitedSpec):
    output_file = File(desc='output file for upload')
    metadata_file = File(desc='metadata file for upload')
    pipeline_config = File(desc='pipeline_config')
    out_files = traits.List(desc='list of files')
    
class pancontainergroup_pan(BaseInterface):
    input_spec = pancontainergroupInputSpec
    output_spec = pancontainergroupOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = pancontainergroup_proc(
            self.inputs.labels_dict,
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='pancontainergroup_node', LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(pancontainergroup_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")

    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    return pan_node





from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
from panpipelines.utils import transformer as tran
import os
import glob
from bids import BIDSLayout
import shlex
import subprocess
from nipype import logging as nlogging
from shutil import copy
import nibabel as nib
import numpy as np
from nilearn.image import resample_to_img

IFLOGGER=nlogging.getLogger('nipype.interface')

def tracula_proc(labels_dict,bids_dir="",tracula_dir=""):

    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')
    params= getParams(labels_dict,"TRACULA_PARAMS")

    layout = BIDSLayout(bids_dir)

    subject="sub-"+participant_label
    session="ses-"+session_label
    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)

    subjects_dir = getParams(labels_dict,'SUBJECTS_DIR')
    os.environ["SUBJECTS_DIR"]=subjects_dir

    command_base, container = getContainer(labels_dict,nodename="tracula", SPECIFIC="TRACULA_CONTAINER",LOGGER=IFLOGGER) 
    command_base = substitute_labels(command_base, labels_dict)

    container_subjects_dir = translate_binding(command_base,subjects_dir)
    os.environ["SINGULARITYENV_SUBJECTS_DIR"]=container_subjects_dir
    
    if not tracula_dir:
        if getParams(labels_dict,'TRACULA_DIR'):
            tracula_dir = getParams(labels_dict,'TRACULA_DIR')
        else:
            tracula_dir = os.path.join(cwd,'tracula_dir')
    if not os.path.isdir(tracula_dir):
        os.makedirs(tracula_dir,exist_ok=True)
    container_tracula_dir=translate_binding(command_base,tracula_dir)
    tracula_config = os.path.join(tracula_dir,f"dmrirc_{subject}_{session}")

    dtroot=os.path.join(tracula_dir,"dtroot")
    if not os.path.isdir(dtroot):
        os.makedirs(dtroot,exist_ok=True)

    dcmroot=os.path.join(tracula_dir,"dcmroot")
    if not os.path.isdir(dcmroot):
        os.makedirs(dcmroot,exist_ok=True)

    workdir=os.path.join(tracula_dir,"workdir")
    if not os.path.isdir(workdir):
        os.makedirs(workdir,exist_ok=True)


    if not os.path.exists(tracula_config) or params == "-prep":
        dwi_entity={}
        dwi_entity["suffix"]="dwi"
        dwi_entity["subject"]=participant_label
        dwi_entity["extension"]='nii.gz'
        if session_label:
            dwi_entity["session"]=session_label
        dwi = layout.get(invalid_filters='allow', **dwi_entity)

        if dwi:
            dwipath=dwi[0].path
            dwibase=os.path.basename(dwipath)
            bvecpath= layout.get_bvec(dwipath)
            bvecbase=os.path.basename(bvecpath)
            bvalpath= layout.get_bval(dwipath)
            bvalbase=os.path.basename(bvalpath)
            dwimeta=dwi[0].get_metadata()
            echospacing = dwimeta["EffectiveEchoSpacing"]
            epifactor = dwimeta["EchoTrainLength"]

            copy(dwipath,os.path.join(dcmroot,dwibase))
            copy(bvecpath,os.path.join(dcmroot,bvecbase))

            with open(bvalpath,"r") as infile:
                bvals=infile.read()
            bvals='\n'.join(bvals.split())
            with open(os.path.join(dcmroot,bvalbase),"w") as outfile:
                outfile.write(bvals)

            dwirpe = [x for x in dwi[0].get_associations() if "dir" in x.path]
            if dwirpe:
                dwirpepath=dwirpe[0].path
                dwirpebase = os.path.basename(dwirpepath)
                copy(dwirpepath,os.path.join(dcmroot,dwirpebase))

                bvecrpebase=os.path.basename(newfile(outputdir=dcmroot,assocfile=dwirpepath,extension="bvec"))
                rpeshape = nib.load(dwirpepath).shape
                dimt=1
                if len(rpeshape) > 3:
                    dimt=rpeshape[3]
                bvecs = np.zeros((3, dimt)) 
                np.savetxt(os.path.join(dcmroot,bvecrpebase), bvecs, fmt="%.3f", delimiter=" ")

                bvalrpebase=os.path.basename(newfile(outputdir=dcmroot,assocfile=dwirpepath,extension="bval"))
                bvals = np.zeros((dimt, 1)) 
                np.savetxt(os.path.join(dcmroot,bvalrpebase), bvals, fmt="%.0f", delimiter=" ")

            else:
                IFLOGGER.error(f"Cannot find DWIRPE for subject. Cannot proceed")
                sys.exit(1)
        else:
            IFLOGGER.error(f"Cannot find DWI for subject. Cannot proceed")
            sys.exit(1)

        config_lines=f"setenv SUBJECTS_DIR {container_subjects_dir}\n" \
                    f"set dtroot = ({container_tracula_dir}/dtroot)\n"\
                    f"set subjlist = ({subject} {subject})\n"\
                    f"set dcmroot = ({container_tracula_dir}/dcmroot)\n"\
                    f"set dcmlist = ({dwibase} {dwirpebase})\n"\
                    f"set bveclist = ({bvecbase} {bvecrpebase})\n"\
                    f"set bvallist = ({bvalbase} {bvalrpebase})\n"\
                    f"set dob0 = 2\n"\
                    f"set pedir = (AP PA)\n"\
                    f"set echospacing = {echospacing*1000}\n"\
                    f"set epifactor = {epifactor}\n"\
                    f"set doeddy = 2 "
                        
        with open(tracula_config,"w") as outfile:
                outfile.write(config_lines)
           
  
    FREEVER="Unknown"
    IFLOGGER.info("Checking the recon-all version:")
    command = f"{command_base} recon-all --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    if "-7.3.2-" in results:
        FREEVER="7.3.2"

    if container:
        IFLOGGER.info("\nChecking the container version:")
        command = f"{container} --version"
        evaluated_command=substitute_labels(command, labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

    if not params == "regacpc":
        command=f"{command_base} {params} -c {tracula_config}"
        evaluated_command=substitute_labels(command,labels_dict)
        results = runCommand(evaluated_command,IFLOGGER)

    if params == "-path" or params == "regacpc":

        command_base, container = getContainer(labels_dict,nodename="regacpc_tracula", SPECIFIC="PAN_CONTAINER",LOGGER=IFLOGGER) 
        command_base = substitute_labels(command_base, labels_dict)

        THRESH = getParams(labels_dict,"TRACULA_THRESH")
        if not THRESH:
            THRESH="20"
        dmri_dir= os.path.join(dtroot,subject,"dmri")
        bvec=os.path.join(dmri_dir,"bvecs")
        bval=os.path.join(dmri_dir,"bvals")
        dwipreproc=os.path.join(dmri_dir,"data.nii.gz")

        b0=newfile(outputdir=workdir,assocfile="b0.nii.gz")
        command=f"{command_base} dwiextract -fslgrad {bvec} {bval} {dwipreproc} -bzero {b0}"
        results = runCommand(command)

        meanb0=newfile(outputdir=workdir,assocfile="meanb0.nii.gz")
        command=f"{command_base} fslmaths {b0} -Tmean {meanb0}"
        results = runCommand(command)

        qsiprep_in=getParams(labels_dict,"QSIPREP_OUTPUT_DIR")
        t1acpc=os.path.join(qsiprep_in,subject,"anat",f"{subject}_desc-preproc_T1w.nii.gz")
        dwiref=os.path.join(qsiprep_in,subject,session,"dwi",f"{subject}_{session}_space-T1w_dwiref.nii.gz")
        t1acpc_mask=os.path.join(qsiprep_in,subject,"anat",f"{subject}_desc-brain_mask.nii.gz")
        t1acpc_brain=os.path.join(workdir,f"{subject}_desc-brain.nii.gz")

        command=f"{command_base} fslmaths {t1acpc} -mas {t1acpc_mask} {t1acpc_brain}"
        results = runCommand(command)

        # resample t1acpc to dwi space
        dwiref_img=nib.load(dwiref)

        t1acpc_img=nib.load(t1acpc)
        t1acpc_dwispace = newfile(outputdir=workdir,assocfile=t1acpc,suffix="space-dwi")
        t1acpc_dwispace_img = resample_to_img(t1acpc_img,dwiref_img)
        nib.save(t1acpc_dwispace_img, t1acpc_dwispace)

        t1acpc_brain_img=nib.load(t1acpc_brain)
        t1acpc_brain_dwispace = newfile(outputdir=workdir,assocfile=t1acpc_brain,suffix="space-dwi")
        t1acpc_brain_dwispace_img = resample_to_img(t1acpc_brain_img,dwiref_img,interpolation='nearest')
        nib.save(t1acpc_brain_dwispace_img, t1acpc_brain_dwispace)

        epi_t1acpc=newfile(outputdir=workdir,assocfile="epi_t1acpc")
        epi_t1acpc_mat=newfile(outputdir=workdir,assocfile=epi_t1acpc,extension="mat")
        command=f"{command_base} epi_reg --epi={meanb0} --t1={t1acpc_dwispace} --t1brain={t1acpc_brain_dwispace} --out={epi_t1acpc}"
        results = runCommand(command)

        dpath = os.path.join(dtroot,subject,"dpath")
        tract_acpc_dir = os.path.join(os.path.dirname(dpath),"dpath_acpc")
        os.makedirs(tract_acpc_dir,exist_ok=True)
        for tractdir in os.listdir(dpath):
            tract = os.path.join(dpath,tractdir,"path.pd.nii.gz")
            tract_acpc = newfile(outputdir=tract_acpc_dir,assocfile=tract,prefix=tractdir,intwix="space-acpc")
            tran.applyAffine_flirt(tract, dwiref, tract_acpc, epi_t1acpc_mat, command_base, interp="nearestneighbour")
            tract_acpc_thresh = newfile(outputdir=tract_acpc_dir,assocfile=tract_acpc,suffix=f"desc-thresh{THRESH}") 
            command=f"{command_base} fslmaths {tract} -thrp {THRESH} -bin {tract_acpc_thresh}"
            results = runCommand(command)
            tran.applyAffine_flirt(tract_acpc_thresh, dwiref, tract_acpc_thresh, epi_t1acpc_mat, command_base, interp="nearestneighbour")


    out_files=[]
    out_files.append(tracula_config)

    return {
        "out_files":out_files
    }



class traculaInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)
    tracula_dir = traits.String("",desc="Tracula Output Directory", usedefault=True)

class traculaOutputSpec(TraitedSpec):
    out_files = traits.List(desc='list of files')
    
class tracula_pan(BaseInterface):
    input_spec = traculaInputSpec
    output_spec = traculaOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = tracula_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='tracula_node',bids_dir="",tracula_dir="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(tracula_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





#!/usr/bin/env python
from panpipelines.utils.util_functions import *
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from functools import partial
import json
import pandas as pd
import os
import logging
import sys
import multiprocessing as mp
from panpipelines.version import __version__
from panpipelines.single_subject import runSingleSubject
from panpipelines.group_subjects import runGroupSubjects
import shutil
import time

LOGGER = logger_setup("panpipelines", logging.DEBUG)
LOGGER.propagate = False
logger_addstdout(LOGGER, logging.INFO)

PROCESSING_OPTIONS=["slurm", "local"]


def parse_params():
    parser = ArgumentParser(description="Pan pipelines")
    PathExists = partial(path_exists, parser=parser)
    parser.add_argument("config_file", type=PathExists, help="Pipeline Configuration File")
    parser.add_argument("--pipeline_outdir", help="directory to place pipeline output")
    parser.add_argument("--participants_file", type=PathExists, help="list of participants")
    parser.add_argument("--sessions_file", type=PathExists, help="Comprehensive list of participants and sessions")
    parser.add_argument("--pipelines", nargs="+")
    parser.add_argument("--pipeline_match", nargs="+")
    parser.add_argument("--pipeline_exclude", nargs="+")
    parser.add_argument("--projects", nargs="+")
    parser.add_argument("--participant_label", nargs="*", type=drop_sub, help="filter by subject label (the sub- prefix can be removed).")
    parser.add_argument("--participant_label_fromfile", type=PathExists, help="participants from file.")    
    parser.add_argument("--incremental", default="False")
    parser.add_argument("--info_delta", default="False", help="what are the subjects that are left to process")
    parser.add_argument("--all_group", default="True")
    parser.add_argument("--force_bids_download",default="False")
    parser.add_argument("--run_dependent_pipelines",nargs="+")
    parser.add_argument("--run_interactive",default="True")
    parser.add_argument("--force_run",default="False")
    parser.add_argument("--participant_query", nargs="+", help="Apply query on pandas. works in exclusion to all other participant query filters")
    parser.add_argument("--participant_exclusions", nargs="*", type=drop_sub, help="filter by subject label (the sub- prefix can be removed).")
    parser.add_argument("--session_label", nargs="*", type=drop_ses, help="filter by session label (the ses- prefix can be removed).")
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))
    return parser


def main():
    datelabel = datetime.datetime.now().strftime("%Y%m%d_%H%M%S_%f")

    job_ids={}

    parser=parse_params()
    args, unknown_args = parser.parse_known_args()

    panpipe_labels={}

    #  TO DO:
    #  In general need to think through how config variables are pulled in and overwritten at the 
    #  general and pipeline level. Current approach leaves a lot to be desired.
    #  We are lucky that pipeline_dir is really only referenced within the pipelines themselves and so we
    #  can get away with this here.
    #
    #  Envisage an approach whereby variables are referenced by a pipeline from both general and pipeline specific
    #  Using keyword in dictionary.
    #
    pipeline_outdir = args.pipeline_outdir
    if pipeline_outdir:
        pipeline_outdir = os.path.abspath(pipeline_outdir)
        panpipe_labels = updateParams(panpipe_labels,"PIPELINE_DIR",pipeline_outdir)

    panpipeconfig_file=str(args.config_file)
    panpipeconfig_json=None
    if os.path.exists(panpipeconfig_file):
        with open(panpipeconfig_file, 'r') as infile:
            panpipeconfig_json = json.load(infile)
    # obtain labels but do not process yet
    panpipe_labels = process_labels(panpipeconfig_json,panpipeconfig_file,panpipe_labels,insert=True, postpone=True)

    if __file__:
        pkgdir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
        panpipe_labels = insertParams(panpipe_labels,'PKG_DIR',pkgdir)
    panpipe_labels = insertParams(panpipe_labels,'PWD',str(os.getcwd()))

   # only need date label if pipeline outdir not provided. This helps with nipype caching for repeat runs
    if not pipeline_outdir:
        pipeline_outdir_suffix = "_" + datelabel
        pipeline_outdir_config = getParams(panpipe_labels,"PIPELINE_DIR")
        if pipeline_outdir_config:
            panpipe_labels = updateParams(panpipe_labels,"PIPELINE_DIR",pipeline_outdir_config + pipeline_outdir_suffix)
        else:
            pipeline_outdir_config = os.path.join(os.getcwd(),"pan_output" +  pipeline_outdir_suffix) 
            panpipe_labels = updateParams(panpipe_labels,"PIPELINE_DIR",pipeline_outdir_config)


    # Now we can resolve all the references based on precedence
    panpipe_labels = update_labels(panpipe_labels)

    pipeline_outdir=os.path.abspath(getParams(panpipe_labels,"PIPELINE_DIR"))
    if not os.path.exists(pipeline_outdir):
        os.makedirs(pipeline_outdir,exist_ok=True)

    # reproducible and slightly more robust name to faciliate nipype caching
    panlabel=os.path.basename(pipeline_outdir)

    LOGFILE=os.path.join(pipeline_outdir,f"{panlabel}_pan_processing.log")
    logger_addfile(LOGGER, LOGFILE, logging.DEBUG)

    LOGGER.info(f"Running {__file__} v{__version__}")
    LOGGER.info("Running Pan Processing")
    LOGGER.info(f"start logging to {LOGFILE}")

    label_key="CONFIG_FILE"
    label_value=panpipeconfig_file
    panpipe_labels = updateParams(panpipe_labels, label_key,label_value)

    xnat_host = getParams(panpipe_labels,"XNAT_HOST")
    credentials = os.path.abspath(getParams(panpipe_labels,"CREDENTIALS"))
    if credentials is not None and os.path.exists(credentials):
        with open(credentials, 'r') as infile:
            cred_dict = json.load(infile)
            cred_user = getParams(cred_dict,"user")
            cred_password = getParams(cred_dict,"password")
    else:
        cred_user = "dummy_user"
        cred_password = "dummy_password"

    SHARED_PROJECT_LIST = getParams(panpipe_labels,"SHARED_PROJECT_LIST")
    if not SHARED_PROJECT_LIST:
        SHARED_PROJECT_LIST=["001_HML","002_HML","003_HML","004_HML"]

    PHANTOM_LIST= getParams(panpipe_labels,"PHANTOM_LIST")
    if not PHANTOM_LIST:
        PHANTOM_LIST=[]

    participants_file = args.participants_file
    if args.participants_file is not None:
        participants_file=str(args.participants_file)
        label_key="PARTICIPANTS_FILE"
        label_value=participants_file
        panpipe_labels = updateParams(panpipe_labels, label_key,label_value)
    else:
        participants_file = getParams(panpipe_labels,"PARTICIPANTS_FILE")

    sessions_file = args.sessions_file
    if args.sessions_file is not None:
        sessions_file=str(args.sessions_file)
        label_key="SESSIONS_FILE"
        label_value=sessions_file
        panpipe_labels = updateParams(panpipe_labels, label_key,label_value)
    else:
        sessions_file = getParams(panpipe_labels,"SESSIONS_FILE")

    if not participants_file:
        participants_file = os.path.join(pipeline_outdir,"xnat_participants_list.tsv")
        panpipe_labels = updateParams(panpipe_labels, "PARTICIPANTS_FILE",participants_file)

    if not sessions_file:
        sessions_file = os.path.join(pipeline_outdir,"xnat_participants_list.tsv")
        panpipe_labels = updateParams(panpipe_labels, "SESSIONS_FILE",sessions_file)

    projects=args.projects
    if not projects:
        projects=["001_HML","002_HML","003_HML","004_HML"]
        LOGGER.info(f"--projects parameter not set. Using defaults {projects} to retrieve subjects and sessions from XNAT.")
    else:
        LOGGER.info(f"Retrieving subjects and sessions from XNAT project {projects}")

    INFO_DELTA = isTrue(args.info_delta)
    ALL_GROUP = isTrue(args.all_group)
    if not ALL_GROUP:
        panpipe_labels = updateParams(panpipe_labels,"ALL_GROUP","N")
    else:
        panpipe_labels = updateParams(panpipe_labels,"ALL_GROUP","Y")       
    INCREMENTAL = isTrue(args.incremental)

    bids_incremental_dir = getParams(panpipe_labels,"BIDS_INCREMENTAL_DIR")
    if not bids_incremental_dir:
        sinkdir = substitute_labels(getParams(panpipe_labels,"SINKDIR_GROUP"),panpipe_labels)
        if not sinkdir:
            derivatives_dir = substitute_labels(getParams(panpipe_labels,"DERIVATIVES_DIR"),panpipe_labels)
            if not derivatives_dir:
                bids_incremental_dir = os.path.join(pipeline_outdir,"derivatives","output_tables","bids_incremental") 
            else:
                bids_incremental_dir = os.path.join(pipeline_outdir,"output_tables","bids_incremental") 
        else:
            bids_incremental_dir = os.path.join(os.path.dirname(sinkdir),"bids_incremental")
    
    if not os.path.exists(bids_incremental_dir):
        os.makedirs(bids_incremental_dir,exist_ok=True)

    if INFO_DELTA:
        new_targetdir=bids_incremental_dir
        info_delta_suffix = f"latest-{datelabel}"
        new_participants_file = newfile(outputdir=new_targetdir, assocfile=participants_file,suffix=info_delta_suffix )

        LOGGER.info(f"Comparing latest participants file on XNAT with current participants file to determine subjects that require downloading.")
        LOGGER.info(f"Downloading latest participants file to {new_targetdir} with suffix of {info_delta_suffix}")
        getBidsTSV(xnat_host,cred_user,cred_password,projects,"BIDS-AACAZ",new_participants_file,demographics=False,shared_project_list=SHARED_PROJECT_LIST,phantom_list=PHANTOM_LIST)
        
        df1 = pd.read_table(new_participants_file,sep="\t")
        participants_list1 = df1["hml_id"].tolist()
        df2 = pd.read_table(participants_file,sep="\t")
        participants_list2 = df2["hml_id"].tolist()
        participant_incremental = [ drop_sub(x) for x in list(set(participants_list1).difference(set(participants_list2)))]
        LOGGER.info(f"Incremental participants found {participant_incremental}")
        mask = df1["hml_id"].isin(participant_incremental)
        incremental_df = df1[mask]
        incremental_tsv = newfile(assocfile=new_participants_file,suffix="delta")
        LOGGER.info(f"Saving delta participant information to {incremental_tsv}.")
        incremental_df.to_csv(incremental_tsv,sep="\t",header=True, index=False)
        panpipe_labels = updateParams(panpipe_labels,"INCREMENTAL_PARTICIPANTS_FILE",incremental_tsv)
        LOGGER.info(f"Quitting.")
        sys.exit(0)


    TEMPLATEFLOW_HOME=getParams(panpipe_labels,"TEMPLATEFLOW_HOME")
    if TEMPLATEFLOW_HOME:
        LOGGER.info(f"Downloading an initial set of TemplateFlow templates for spaces MNI152NLin2009cAsym and MNI152NLin6Asym to {TEMPLATEFLOW_HOME}.")
        initTemplateFlow(TEMPLATEFLOW_HOME)

    RUN_INTERACTIVE=isTrue(args.run_interactive)
    dependent_pipelines = args.run_dependent_pipelines
    if dependent_pipelines:
        RUN_DEPENDENT_PIPELINES = True
    else:
        RUN_DEPENDENT_PIPELINES = False
        dependent_pipelines=[]

    FORCE_BIDS_DOWNLOAD = isTrue(args.force_bids_download)
    if FORCE_BIDS_DOWNLOAD:
        panpipe_labels = updateParams(panpipe_labels,"FORCE_BIDS_DOWNLOAD","Y")
        LOGGER.info(f"FORCE_BIDS_DOWNLOAD set to Y")

    FORCE_RUN = isTrue(args.force_run)
    if FORCE_RUN:
        panpipe_labels = updateParams(panpipe_labels,"FORCE_RUN","Y")
        LOGGER.info(f"FORCE_RUN set to Y. All selected pipelines will be forced to rerun")

    if INCREMENTAL:
        LOGGER.info("Running panprocessing in incremental mode.")
        if not os.path.exists(participants_file):
            LOGGER.info(f"Participants file {participants_file} not present. Cannot run processing in incremental mode")
            INCREMENTAL= False
        else:
            old_participants_file=newfile(assocfile=participants_file,suffix=datelabel)
            shutil.move(participants_file,old_participants_file)
            old_sessions_file=newfile(assocfile=sessions_file,suffix=datelabel)
            shutil.move(sessions_file,old_sessions_file)
            bids_participant_file=os.path.join(getParams(panpipe_labels,"BIDS_DIR"),"participants.tsv")
            backup_participant_file= newfile(assocfile=bids_participant_file,suffix=datelabel)
            shutil.copy(bids_participant_file,backup_participant_file)

    # if participants file doesn't exist then lets download it
    if not os.path.exists(sessions_file):
        if not os.path.exists(participants_file):
            LOGGER.info(f"Participants file not found at {participants_file} - retrieving from XNAT. Please wait.")
            targetdir = os.path.dirname(participants_file)
            if not os.path.exists(targetdir):
                os.makedirs(targetdir,exist_ok=True)

        if not os.path.exists(sessions_file):
            LOGGER.info(f"Sessions file not found at {sessions_file} - retrieving from XNAT. Please wait.")
            targetdir = os.path.dirname(sessions_file)
            if not os.path.exists(targetdir):
                os.makedirs(targetdir,exist_ok=True)

        getBidsTSV(xnat_host,cred_user,cred_password,projects,"BIDS-AACAZ",participants_file,demographics=False,shared_project_list=SHARED_PROJECT_LIST,phantom_list=PHANTOM_LIST,sessionsTSV=sessions_file)


    pipelines=args.pipelines
    if args.pipelines is not None:
        label_key="PIPELINES"
        label_value=pipelines
        panpipe_labels = updateParams(panpipe_labels, label_key,label_value)
    else:
        pipelines = getParams(panpipe_labels,"PIPELINES")

    if not pipelines:
        pipelines=[]

    pipeline_match=args.pipeline_match
    pipeline_exclude=args.pipeline_exclude

    ALL_PIPELINES=[p for p in panpipeconfig_json.keys() if p != "all_pipelines"]
    if not pipelines and not dependent_pipelines:
        LOGGER.info("No pipelines specified at command line.")
        pipelines = ALL_PIPELINES
        if pipeline_match:
            LOGGER.info(f"All pipelines matching {str(pipeline_match)} in configuration file will be run.")
            pipeline_select=[]
            for pipe_match in pipeline_match:
                pipeline_select.extend([p for p in pipelines if pipe_match in p])
            pipelines = pipeline_select
        else:
            LOGGER.info("All pipelines in configuration file will be run.")

    if RUN_DEPENDENT_PIPELINES:
        pipelines.extend(get_dependent_pipelines(panpipeconfig_json,dependent_pipelines,ALL_PIPELINES))
        panpipe_labels = updateParams(panpipe_labels, "RUN_DEPENDENT_PIPELINES","Y")

    if pipeline_exclude and pipelines:
        LOGGER.info(f"Excluding pipelines {pipeline_exclude} from this run.")
        pipelines = list(set(pipelines) - set(pipeline_exclude))

    # Remove duplicates in pipeline list
    pipelines = list(set(pipelines))

    LOGGER.info(f"About to arrange pipelines by dependency. Pipeline list is {pipelines}")    
    pipelines = arrangePipelines(panpipeconfig_json,pipelines=pipelines)

    LOGGER.info(f"Pipelines arranged by dependency. Pipeline list is {pipelines}")    

    participant_label = args.participant_label
    if args.participant_label is not None:
        label_key="PARTICIPANTS"
        label_value=participant_label
        panpipe_labels = updateParams(panpipe_labels, label_key,label_value)
    else:
        participant_label = getParams(panpipe_labels,"PARTICIPANTS")

    participant_label_fromfile = args.participant_label_fromfile
    if participant_label_fromfile:
        with open(participant_label_fromfile, 'r') as infile:
            parts = infile.read()
        split_parts = parts.split("\n")
        split_parts_valid = [x for x in split_parts if x]

        if split_parts_valid:
            label_key="PARTICIPANTS"
            participant_label = split_parts_valid
            label_value=participant_label
            panpipe_labels = updateParams(panpipe_labels, label_key,label_value)


    participant_query = args.participant_query
    if not participant_query:
        participant_query = ""
    else:
        participant_query = " ".join(participant_query)

    participant_exclusions = args.participant_exclusions
    if not participant_exclusions:
        participant_exclusions = []
    
    session_label = args.session_label
    if args.session_label is not None:
        label_key="SESSION_LABEL"
        label_value=session_label
        panpipe_labels = updateParams(panpipe_labels, label_key,label_value)
    else:
        session_label = getParams(panpipe_labels,"SESSION_LABEL")
        if not session_label:
            session_label="ALL_SESSIONS"
            panpipe_labels = updateParams(panpipe_labels, "SESSION_LABEL",session_label)

    if INCREMENTAL:
        LOGGER.info("Running in incremental mode. Identifying participants to run")
        df1 = pd.read_table(participants_file,sep="\t")
        participants_list1 = df1["hml_id"].tolist()
        df2 = pd.read_table(old_participants_file,sep="\t")
        participants_list2 = df2["hml_id"].tolist()
        participant_incremental = [ drop_sub(x) for x in list(set(participants_list1).difference(set(participants_list2)))]
        LOGGER.info(f"Incremental participants found {participant_incremental}")

        new_targetdir=bids_incremental_dir
        info_delta_suffix = f"latest-{datelabel}"
        mask = df1["hml_id"].isin(participant_incremental)
        incremental_df = df1[mask]
        incremental_tsv = newfile(outputdir= new_targetdir, assocfile=participants_file,suffix=info_delta_suffix + "-" + "delta")
        LOGGER.info(f"Saving delta participant information to {incremental_tsv}.")
        incremental_df.to_csv(incremental_tsv,sep="\t",header=True, index=False)
        panpipe_labels = updateParams(panpipe_labels,"INCREMENTAL_PARTICIPANTS_FILE",incremental_tsv)

        participant_label.extend(participant_incremental)
        LOGGER.info(f"Participants to process {participant_label}")

    # take snapshot of the runtime labels for all pipelines
    runtime_labels = panpipe_labels.copy()

    orig_participant_label = participant_label

    if INCREMENTAL:
        if not pipelines:
            LOGGER.info(f"Running in incremental mode but no pipelines selected. Will run the dummy_panpipeline and ftp_upload_pan script if they are not explictly excluded")
            if "dummy" not in pipeline_exclude:
                pipelines.add("dummy")
            
            if "ftp_upload_bids" not in pipeline_exclude:
                pipelines.add("ftp_upload_bids")

    if not participant_query:            
        LOGGER.info(f"participants to be processed:\n {participant_label}.\n Note that subject exclusions will be applied at the pipeline level.\n")
    else:
        projectmap = get_projectmap_query(sessions_file,participant_query)
        participant_list = projectmap[0]
        LOGGER.info(f"participants to be processed:\n {participant_list}.\n Note that subject exclusions will be applied at the pipeline level.\n")
    time.sleep(1)
    LOGGER.info(f"Pipelines to be processed :\n {pipelines}\n")
    time.sleep(1)

    key=""
    if RUN_INTERACTIVE:
        key=input("press c to continue, q to exit")
    while RUN_INTERACTIVE and not key.upper() == "C":
        if key.upper() == "Q":
            sys.exit()
        key=input()
        
    for pipeline in pipelines:

        subject_exclusions=[]
        subject_exclusions.extend(participant_exclusions)

        pipeline_panpipe_labels={}
        pipeline_panpipe_labels = process_labels(panpipeconfig_json,panpipeconfig_file,pipeline_panpipe_labels,pipeline)

        # panpipe_labels holds global defs - need a more elegant way to handle pipeline and global defs
        participant_override_file = substitute_labels(getParams(pipeline_panpipe_labels,"PARTICIPANT_OVERRIDE_FILE"),panpipe_labels)
        if participant_override_file:
            with open(participant_override_file, 'r') as infile:
                parts = infile.read()
            split_parts = parts.split("\n")
            split_parts_valid = [x for x in split_parts if x]

            if split_parts_valid:
                participant_label = split_parts_valid
                panpipe_labels = updateParams(panpipe_labels, "PARTICIPANTS",participant_label)
                LOGGER.info(f"Participant override has been performed at pipeline level :\n {participant_label} will be run instead for {pipeline}\n")
        else:
            participant_label = orig_participant_label
            panpipe_labels = updateParams(panpipe_labels, "PARTICIPANTS",participant_label)

        
        pipeline_exclusions = getParams(pipeline_panpipe_labels,"PARTICIPANT_PIPELINE_EXCLUSIONS")
        if not pipeline_exclusions:
            pipeline_exclusions=[]
        subject_exclusions.extend(pipeline_exclusions) 
        panpipe_labels = updateParams(panpipe_labels,"EXCLUDED_PARTICIPANTS",subject_exclusions)

        if not participant_query:
            projectmap = get_projectmap(participant_label, participants_file,session_labels=session_label,sessions_file=sessions_file,subject_exclusions=subject_exclusions)
        else:
            projectmap = get_projectmap_query(sessions_file,participant_query,subject_exclusions=subject_exclusions)
        participant_list = projectmap[0]
        panpipe_labels = updateParams(panpipe_labels, "PARTICIPANTS",participant_list)
        project_list  = projectmap[1]
        session_list = projectmap[2]
        shared_project_list  = projectmap[3]

        # obtain mappings for all subjects which will be handy for incremental
        projectmap_all = get_projectmap(["ALL_SUBJECTS"],participants_file,session_labels=session_label,sessions_file=sessions_file,subject_exclusions=subject_exclusions)
        participant_list_all = projectmap_all[0]
        project_list_all  = projectmap_all[1]
        session_list_all = projectmap_all[2]
        shared_project_list_all = projectmap_all[3]

        LOGGER.info(f"Processing pipeline : {pipeline}")
        panpipe_labels = updateParams(panpipe_labels, "PIPELINE", pipeline)
        panpipe_labels = process_labels(panpipeconfig_json,panpipeconfig_file,panpipe_labels,pipeline)

        ALL_GROUP = isTrue(getParams(panpipe_labels,"ALL_GROUP"))

        analysis_level = getParams(panpipe_labels,"ANALYSIS_LEVEL")
        if analysis_level == "group":
            # rather rigid - but we foreshadow the worfflow directory here and use a param so it is useable
            wf_dir = f"<PIPELINE_DIR>/{pipeline}/group/{pipeline}_wf"
            panpipe_labels = updateParams(panpipe_labels,f"WORKFLOW_DIR",wf_dir)
        else:
            wf_dir = f"<PIPELINE_DIR>/{pipeline}/<PARTICIPANT_XNAT_PROJECT>/sub-<PARTICIPANT_LABEL>/ses-<PARTICIPANT_SESSION>/{pipeline}_wf"
            panpipe_labels = updateParams(panpipe_labels,f"WORKFLOW_DIR",wf_dir)

        # We handle the <DEPENDENCY> key specially as this is a list that we need to resolve into DEPENDENCY1, DEPENDENCY2, ...DEPENDENCYN
        dependency_list = getParams(panpipe_labels,"DEPENDENCY")
        if dependency_list:
            if not isinstance(dependency_list,list):
                dependency_list=[dependency_list]
                # can refer to single DEPENDENCY also as DEPENDENCY1
                dependency_dir = f"<PIPELINE_DIR>/<DEPENDENCY>/<PARTICIPANT_XNAT_PROJECT>/sub-<PARTICIPANT_LABEL>/ses-<PARTICIPANT_SESSION>/<DEPENDENCY>_wf"
                panpipe_labels = updateParams(panpipe_labels,f"DEPENDENCY_DIR",dependency_dir)

            depcount=1

            for dependency in dependency_list:
                panpipe_labels = updateParams(panpipe_labels,f"DEPENDENCY{depcount}",dependency)
                dependency_dir = f"<PIPELINE_DIR>/<DEPENDENCY{depcount}>/<PARTICIPANT_XNAT_PROJECT>/sub-<PARTICIPANT_LABEL>/ses-<PARTICIPANT_SESSION>/<DEPENDENCY{depcount}>_wf"
                panpipe_labels = updateParams(panpipe_labels,f"DEPENDENCY{depcount}_DIR",dependency_dir)
                depcount = depcount + 1
        
        # Now we can resolve all the references based on precedence in the pipeline section
        panpipe_labels = update_labels(panpipe_labels)

        processing_environment=getParams(panpipe_labels,"PROCESSING_ENVIRONMENT") 

        if processing_environment is None  or processing_environment not in PROCESSING_OPTIONS:
            processing_environment = "slurm"

        if processing_environment == "slurm":
            analysis_node = getParams(panpipe_labels,"ANALYSIS_NODE")

            if analysis_level == "group":
                updateParams(panpipe_labels, "SLURM_TEMPLATE", getParams(panpipe_labels,"SLURM_GROUP_TEMPLATE"))
                if participant_list and not ALL_GROUP:
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_LABEL",participant_list)
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_PROJECT",project_list)
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_SHARED_PROJECT",shared_project_list)
                    updateParams(panpipe_labels,"GROUP_SESSION_LABEL",session_list)
                else:
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_LABEL",participant_list_all)
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_PROJECT",project_list_all)
                    updateParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_SHARED_PROJECT",shared_project_list_all)
                    updateParams(panpipe_labels,"GROUP_SESSION_LABEL",session_list_all)

            else:
                updateParams(panpipe_labels, "SLURM_TEMPLATE", getParams(panpipe_labels,"SLURM_PARTICIPANT_TEMPLATE"))

            if analysis_node == "gpu":
                updateParams(panpipe_labels, "SLURM_HEADER", getParams(panpipe_labels,"SLURM_GPU_HEADER"))
            else:
                updateParams(panpipe_labels, "SLURM_HEADER", getParams(panpipe_labels,"SLURM_CPU_HEADER"))

            try:
                jobid = submit_script(participant_list, participants_file, pipeline, panpipe_labels,job_ids, analysis_level,projects_list = project_list, sessions_list=session_list, sessions_file = sessions_file, LOGGER=LOGGER,panlabel=panlabel)

                job_ids[pipeline]=jobid
            except Exception as ex:
                LOGGER.info(f"problems running pipeline {pipeline}. Details: {ex}")

        elif processing_environment == "local":
            pipeline_class = getParams(panpipe_labels,"PIPELINE_CLASS")
            if pipeline_class is None:
                pipeline_class = pipeline 

            #use_pipeline_desc = getParams(panpipe_labels,"USE_PIPELINE_DESC")
            #if use_pipeline_desc is None:
            #    use_pipeline_desc = "N"

            #pipeline_desc = getParams(panpipe_labels,"PIPELINE_DESC")
            #if pipeline_desc is None or use_pipeline_desc == "N":
            #    pipeline_desc = pipeline
            #else:
            #    pipeline_desc = "".join([x if x.isalnum() else "_" for x in pipeline_desc])  

            pipeline_outdir=os.path.join(getParams(panpipe_labels,"PIPELINE_DIR"),pipeline)
            if not os.path.exists(pipeline_outdir):
                os.makedirs(pipeline_outdir,exist_ok=True)

            lock_dir=getParams(panpipe_labels,"LOCK_DIR")
            if lock_dir:
                if not os.path.exists(lock_dir):
                    os.makedirs(lock_dir,exist_ok=True)
            else:
                lock_dir=os.path.join(getParams(panpipe_labels,"PIPELINE_DIR"),"datalocks")
                if not os.path.exists(lock_dir):
                    os.makedirs(lock_dir,exist_ok=True)
                panpipe_labels = updateParams(panpipe_labels,"LOCK_DIR",lock_dir)

            bids_dir = getParams(panpipe_labels,"BIDS_DIR")

            execution_json=getParams(panpipe_labels,"NIPYPE_CONFIG")
            if execution_json is None:
                execution_json = {} 
            
            try:
                procs = getProcNums(panpipe_labels)

                if analysis_level == "participant":
                    # run serially if procs < 2
                    if procs > 1:
                        multiproc_zip = zip(participant_list,project_list,shared_project_list,session_list)
                        parrunSingleSubject=partial(runSingleSubject, pipeline=pipeline, pipeline_class=pipeline_class, pipeline_outdir=pipeline_outdir, panpipe_labels=panpipe_labels,bids_dir=bids_dir,cred_user=cred_user,cred_password=cred_password, execution_json=execution_json,analysis_level=analysis_level)
                        with mp.Pool(procs) as pool:
                            completedlist = pool.starmap(parrunSingleSubject,multiproc_zip)
                            pool.close()
                            pool.join()
                    else:
                        for part_count in range(len(participant_list)):
                            if session_list:
                                session_label= session_list[part_count]
                            else:
                                session_label=None
                            runSingleSubject(participant_list[part_count], project_list[part_count], shared_project_list[part_count],session_label,pipeline=pipeline, pipeline_class=pipeline_class, pipeline_outdir=pipeline_outdir, panpipe_labels=panpipe_labels,bids_dir=bids_dir,cred_user=cred_user,cred_password=cred_password, execution_json=execution_json,analysis_level=analysis_level)
                else:
                    if ALL_GROUP:
                        updateParams(panpipe_labels,"GROUP_PARTICIPANTS_LABEL",participant_list_all)
                        updateParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_PROJECT",project_list_all)
                        updateParams(panpipe_labels,"GROUP_SESSION_LABEL",session_list_all)
                        updateParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_SHARED_PROJECT",shared_project_list_all)
                        runGroupSubjects(participant_list_all, project_list_all,shared_project_list_all,session_list_all,pipeline=pipeline, pipeline_class=pipeline_class, pipeline_outdir=pipeline_outdir, panpipe_labels=panpipe_labels,bids_dir=bids_dir,cred_user=cred_user,cred_password=cred_password, execution_json=execution_json,analysis_level=analysis_level)
                    else:
                        updateParams(panpipe_labels,"GROUP_PARTICIPANTS_LABEL",participant_list)
                        updateParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_PROJECT",project_list)
                        updateParams(panpipe_labels,"GROUP_SESSION_LABEL",session_list)
                        updateParams(panpipe_labels,"GROUP_PARTICIPANTS_XNAT_SHARED_PROJECT",shared_project_list)
                        runGroupSubjects(participant_list, project_list,shared_project_list,session_list,pipeline=pipeline, pipeline_class=pipeline_class, pipeline_outdir=pipeline_outdir, panpipe_labels=panpipe_labels,bids_dir=bids_dir,cred_user=cred_user,cred_password=cred_password, execution_json=execution_json,analysis_level=analysis_level)


            except Exception as ex:
                LOGGER.info(f"problems running pipeline {pipeline}. Details: {ex}")
 
        else:
            LOGGER.info(f"processing environment {processing_environment} not currently supported. Options are {PROCESSING_OPTIONS}.")

        # clear out and restore runtime labels for next pipeline
        panpipe_labels = {}
        panpipe_labels = add_labels(runtime_labels,panpipe_labels)
       
    LOGGER.info(f"All pipelines {pipelines} successfully run/submitted.")
    LOGGER.info(f"View logs at {LOGFILE}.")
    LOGGER.debug("\n------  environment dump -----------")
    LOGGER.debug(os.environ)


# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()


import neuroCombat
from panpipelines.utils.util_functions import drop_ses,updateParams,getParams,getGlob,getContainer,substitute_labels
import numpy as np
import pandas as pd
import glob
import os
import nibabel
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from pathlib import Path
from functools import partial
import os
from nipype import logging as nlogging
import json

IFLOGGER=nlogging.getLogger('nipype.interface')
   
def _path_exists(path, parser):
    """Ensure a given path exists."""
    if path is None or not Path(path).exists():
        raise parser.error(f"Path does not exist: <{path}>.")
    return Path(path).expanduser().absolute()

def parse_params():
    parser = ArgumentParser(description="additional_transforms")
    PathExists = partial(_path_exists, parser=parser)
    parser.add_argument("pipeline_config_file", type=PathExists, help="Pipeline Config File")
    return parser


def run_combat(pipeline_config_file):
  
    panpipeconfig_file=str(pipeline_config_file)
    labels_dict =None
    if os.path.exists(pipeline_config_file):
        IFLOGGER.info(f"{pipeline_config_file} exists.")
        with open(pipeline_config_file,'r') as infile:
            labels_dict = json.load(infile)
    
    try:
        basil_table = getGlob(getParams(labels_dict,"BASIL_TABLE"))
        basil_df = pd.read_csv(basil_table,sep=",")
        demo_table = getGlob(getParams(labels_dict,"DEMO_TABLE"))
        demo_df = pd.read_csv(demo_table,sep=",")
        omnibus_df = pd.merge(basil_df, demo_df,  how='left', left_on=["hml_id"], right_on =["hml_id"])

        mask=getParams(labels_dict,"BASIL_MASK")
        maskimg = nibabel.load(mask)
        maskdata = maskimg.get_fdata()
        (dimxx,dimyy,dimzz) = maskdata.shape

        #participants_label = getParams(labels_dict,'GROUP_PARTICIPANTS_LABEL')
        #participants_project = getParams(labels_dict,'GROUP_PARTICIPANTS_XNAT_PROJECT')
        #participants_session = getParams(labels_dict,'GROUP_SESSION_LABEL')
        #for part_vals in zip(participants_label,participants_project,participants_session):
        cbfrefimg = None
        basil_ref = getParams(labels_dict,"BASIL_DATAREF")
        data_masked=np.zeros((1,1))
        for dfnum in range(len(basil_df)):
            hmlid = basil_df.iloc[dfnum].hml_id
            project = basil_df.iloc[dfnum].site_id
            session = drop_ses(basil_df.iloc[dfnum].session_id)
            labels_dict = updateParams(labels_dict,"PARTICIPANT_LABEL",hmlid)
            labels_dict = updateParams(labels_dict,"PARTICIPANT_XNAT_PROJECT",project)
            labels_dict = updateParams(labels_dict,"PARTICIPANT_SESSION",session)
            cbf=substitute_labels(basil_ref,labels_dict)
            if os.path.exists(cbf):
                cbfimg = nibabel.load(cbf)
                if dfnum == 0:
                    cbfrefimg = cbfimg 
                cbfdata = cbfimg.get_fdata()
                cbf_flat = cbfdata[maskdata > 0]
                cbf_flat_valid = np.zeros((len(cbf_flat),1))
                cbf_flat_valid[:,0]=cbf_flat
                if len(data_masked)>1:
                    data_masked = np.hstack((data_masked,cbf_flat_valid))
                else:
                    data_masked = cbf_flat_valid

        wf_dir = getParams(labels_dict,"WORKFLOW_DIR")
        output_dir = os.path.join(wf_dir,"combat_output")
        if not os.path.exists(output_dir):
            os.makedirs(output_dir,exist_ok=True)

        #if len(data_masked) > 1:
        #    np.savetxt(os.path.join(output_dir,'cbfdata.txt'), data_masked, fmt='%.18e', #delimiter=',')
  
        omnibus_df["batch"] = omnibus_df["site_id"].apply(lambda row: int(row.split("_")[0]))   
        batch_col =["batch"]

        health_medical_columns = [x for x in demo_df.columns if x.startswith("health_medical")]

        # To specify names of the variables that are categorical:
        #categorical_cols=["sex","handedness","race","hispanic_latino","highest_education_level_completed","memory_rank"] + health_medical_columns
        categorical_cols=["sex","handedness","race"]
        continuous_cols = ["age"]

        # Convert categorical columns to ordinal values, for specified columns
        covars_df = omnibus_df[["hml_id","subject_id","session_id","site_id"] + batch_col + categorical_cols + continuous_cols]

        # correct Nans
        covars_df.loc[:, categorical_cols] = covars_df.loc[:, categorical_cols].fillna("MISSING")
        covars_df.loc[:, continuous_cols] = covars_df.loc[:, continuous_cols].fillna(-1)

        covars_df = covars_df.apply(lambda col: pd.factorize(col)[0] + 1 if col.name in categorical_cols and col.dtype == 'object' else col)

        #Harmonization step:
        datamask_combat = neuroCombat.neuroCombat(dat=data_masked,
            covars=covars_df,
            batch_col=batch_col,
            continuous_cols=continuous_cols,
            categorical_cols=categorical_cols,
            ref_batch=3)
        harmonized = datamask_combat["data"]
        IFLOGGER.info("1st stage of harmonization completed")

        if not cbfrefimg:
            cbfrefimg = maskimg
        suffix = getParams(labels_dict,"COMBAT_SUFFIX")
        if not suffix:
            suffix="space-MNI152NLin6Asym_desc-combat_cbf"
        for dfnum in range(len(covars_df)):
            subject = covars_df.iloc[dfnum].subject_id
            project = covars_df.iloc[dfnum].site_id
            session = covars_df.iloc[dfnum].session_id
            print(f"processing {dfnum}: {subject} {session} {project}")
            cbf_template=np.zeros((dimxx,dimyy,dimzz))
            cbf_template[maskdata > 0]= harmonized[:,dfnum]
            cbf_combat_img = nibabel.Nifti1Image(cbf_template,cbfrefimg.affine,cbfrefimg.header)
            cbf_combat_file = os.path.join(output_dir,project,subject,session,f"{subject}_{session}_{suffix}.nii.gz")
            os.makedirs(os.path.dirname(cbf_combat_file),exist_ok=True)
            nibabel.save(cbf_combat_img,cbf_combat_file)

    except Exception as e:
        message = f"Exception caught: \n{e}"
        IFLOGGER.error(message)


def main():
    parser=parse_params()
    args, unknown_args = parser.parse_known_args()
    pipeline_config_file = str(args.pipeline_config_file)

    run_combat(pipeline_config_file)

# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()




__all__ = [
    "utils",
    "Factory",
    "pipelines",
    "scripts",
    "workflows",
    "nodes",
    "pan_processing",
    "single_subject",
    "group_subjects",
    "version"
]

from . import version
__version__ = version.__version__


from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
import shlex
import subprocess
from nipype import logging as nlogging
from bids import BIDSLayout

IFLOGGER=nlogging.getLogger('nipype.interface')

IS_PRESENT="^^^"
IGNORE="###"

def mriqc_proc(labels_dict,bids_dir=""):

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)

    command_base, container = getContainer(labels_dict,nodename="mriqc", SPECIFIC="MRIQC_CONTAINER",LOGGER=IFLOGGER)
    TEMPLATEFLOW_HOME=getParams(labels_dict,"TEMPLATEFLOW_HOME")
    os.environ["TEMPLATEFLOW_HOME"]=TEMPLATEFLOW_HOME
    os.environ["SINGULARITYENV_TEMPLATEFLOW_HOME"]=translate_binding(command_base,TEMPLATEFLOW_HOME)

    IFLOGGER.info("Checking the mriqc version:")
    command = f"{command_base} --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    participant_session = getParams(labels_dict,'PARTICIPANT_SESSION')

    # handle field maps - use dwirpe as standard if valid, otherwise use megre
    if not bids_dir:
        bids_dir = substitute_labels('<BIDS_DIR>', labels_dict)
    layout = BIDSLayout(bids_dir)

    bids_filter_dict={}
    bids_filter_dict["bold"] = {}
    bids_filter_dict["bold"]["session"] =  participant_session
    bids_filter_dict["dwi"] = {}
    bids_filter_dict["dwi"]["session"] =  participant_session
    bids_filter_dict["t1w"] = {}
    bids_filter_dict["t1w"]["session"] =  participant_session
    bids_filter_dict["t2w"] = {}
    bids_filter_dict["t2w"]["session"] =  participant_session

    bids_filter_file = os.path.join(cwd,f"{participant_label}_{participant_session}_bids_filter_file.json")
    export_labels(bids_filter_dict,bids_filter_file)
    IFLOGGER.info(f"Specifying session filter: exporting {bids_filter_dict} to {bids_filter_file}")

    mriqc_dict={}
    mriqc_dict = updateParams(mriqc_dict,"--participant_label","<PARTICIPANT_LABEL>")
    mriqc_dict = updateParams(mriqc_dict,"--bids-filter-file",bids_filter_file)
    mriqc_dict = updateParams(mriqc_dict,"--mem","<BIDSAPP_MEMORY>")
    mriqc_dict = updateParams(mriqc_dict,"--nprocs","<BIDSAPP_THREADS>")
    mriqc_dict = updateParams(mriqc_dict,"--omp-nthreads","4")
    mriqc_dict = updateParams(mriqc_dict,"--no-sub",IS_PRESENT)
    mriqc_dict = updateParams(mriqc_dict,"-w","<CWD>/mriqcwork")

    # Additional params
    MRIQC_OVERRIDE_PARAMS = getParams(labels_dict,"MRIQC_OVERRIDE_PARAMS")
    if MRIQC_OVERRIDE_PARAMS and isinstance(MRIQC_OVERRIDE_PARAMS,dict):
        add_labels(MRIQC_OVERRIDE_PARAMS,mriqc_dict)

    # Additional params for specific subjects
    UNIQUE_MRIQC_OVERRIDE_PARAMS = getParams(labels_dict,"UNIQUE_MRIQC_OVERRIDE_PARAMS")
    if UNIQUE_MRIQC_OVERRIDE_PARAMS and isinstance(UNIQUE_MRIQC_OVERRIDE_PARAMS,list):
        for override_definition in UNIQUE_MRIQC_OVERRIDE_PARAMS:
            if isinstance(override_definition,dict) and "CANDIDATES" in override_definition.keys() and "PARAMS" in override_definition.keys():
                subject_list = override_definition["CANDIDATES"]
                subject_params = override_definition["PARAMS"]
                if f"{participant_label}_{participant_session}" in subject_list or f"{participant_label}" in subject_list:
                    add_labels(subject_params,mriqc_dict)        

    params = ""
    for mriqc_tag, mriqc_value in mriqc_dict.items():
        if "--" in mriqc_tag and "---" not in mriqc_tag:
            if mriqc_value == IS_PRESENT:
                params=params + " " + mriqc_tag
            elif mriqc_value == IGNORE:
                IFLOGGER.info(f"Parameter {mriqc_tag} is being skipped. This has been explicitly required in configuration.")
            else:
                # we dont need = sign for mriqc just for basi;
                params = params + " " + mriqc_tag + " " + mriqc_value

        elif "-" in mriqc_tag and "--" not in mriqc_tag:
            params = params + " " + mriqc_tag + " " + mriqc_value

        else:
            print(f"mriqc tag {mriqc_tag} not valid.")

    command=f"{command_base}"\
            " "+ bids_dir +\
            " <CWD>/mriqcout"\
            " participant"\
            " "+ params 

    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)

    bold_rep = getGlob(os.path.join(cwd,'mriqcoutput','sub-{}'.format(participant_label),'ses-*','func','*bold.json'))
    t1w_rep = getGlob(os.path.join(cwd,'mriqcoutput','sub-{}'.format(participant_label),'ses-*','anat','*T1w.json'))
    t2w_rep= getGlob(os.path.join(cwd,'mriqcoutput','sub-{}'.format(participant_label),'ses-*','anat','*T2w.json'))
    dwi_rep = getGlob(os.path.join(cwd,'mriqcoutput','sub-{}'.format(participant_label),'ses-*','dwi','*dwi.json'))

    output_dir = cwd

    out_files=[]
    if bold_rep:
        out_files.insert(0,bold_rep)
    if t1w_rep:
        out_files.insert(0,t1w_rep)
    if t2w_rep:
        out_files.insert(0,t2w_rep)
    if dwi_rep:
        out_files.insert(0,dwi_rep)


    return {
        "output_dir":output_dir,
        "out_files":out_files
    }



class mriqcInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    bids_dir = traits.String("",desc="BIDS Directory", usedefault=True)

class mriqcOutputSpec(TraitedSpec):
    output_dir = traits.String(desc="MRIQC output directory")
    out_files = traits.List(desc='list of files')
    
class mriqc_pan(BaseInterface):
    input_spec = mriqcInputSpec
    output_spec = mriqcOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = mriqc_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='mriqc_node',bids_dir="",LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(mriqc_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if bids_dir is None or bids_dir == "":
        bids_dir = substitute_labels("<BIDS_DIR>", labels_dict)

    pan_node.inputs.bids_dir =  bids_dir

    return pan_node





from nipype import Workflow,Node
from nipype.interfaces.io import DataSink

import panpipelines.nodes.upload as upload
from panpipelines.utils.util_functions import *


def create(name, wf_base_dir,labels_dict,createGraph=True,execution={},LOGGER=None):
    # Create workflow
    pan_workflow = Workflow(name=name, base_dir=wf_base_dir)

    if LOGGER:
        LOGGER.info(f"Created Workflow {name} with base directory {wf_base_dir}")

    if len(execution.keys()) > 0:
        pan_workflow.config = process_dict(pan_workflow.config,execution)

    # Specify node inputs
    source_path = getParams(labels_dict,"FTP_SOURCEPATH")
    source_list=[]
    if isinstance(source_path,list):
        for source_template in source_path:
            evaluated_source_template = substitute_labels(source_template,labels_dict)
            source_files=glob.glob(evaluated_source_template)
            source_list.extend(source_files)
    else:
        source_list=[source_path]

    remote_path = getParams(labels_dict,"FTP_REMOTEPATH")
    remote_list=[]
    if isinstance(remote_path,list):
        for remote_template in remote_path:
            evaluated_remote_template = substitute_labels(remote_template,labels_dict)
            remote_files=evaluated_remote_template
            remote_list.append(remote_files)
    else:
        remote_list=[remote_path]

    ftpcredentials = getParams(labels_dict,"FTPCREDENTIALS")

    upload_node = upload.create(labels_dict,source_list,remote_list, ftpcredentials,LOGGER=LOGGER)

    sinker_dir = getParams(labels_dict,"SINKDIR_GROUP")
    if sinker_dir:
        sinker = Node(DataSink(),name='collatecsvgroup_sink')
        sinker_basedir = os.path.dirname(sinker_dir)
        sinker_folder = os.path.basename(sinker_dir)
        if not os.path.exists(sinker_basedir):
            os.makedirs(sinker_basedir,exist_ok=True)
        sinker.inputs.base_directory = sinker_basedir

        pan_workflow.connect( upload_node,"metadata_file",sinker,f"{sinker_folder}.@metadata_file")
    else:
        pan_workflow.add_nodes([upload_node])


    if createGraph:
         pan_workflow.write_graph(graph2use='flat')


    return pan_workflow



from panpipelines.utils.util_functions import getParams,getGlob,getContainer,substitute_labels
from panpipelines.utils.transformer import ants_registration_rigid
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from pathlib import Path
from functools import partial
import os
from nipype import logging as nlogging
import json

IFLOGGER=nlogging.getLogger('nipype.interface')

def _path_exists(path, parser):
    """Ensure a given path exists."""
    if path is None or not Path(path).exists():
        raise parser.error(f"Path does not exist: <{path}>.")
    return Path(path).expanduser().absolute()

def parse_params():
    parser = ArgumentParser(description="additional_transforms")
    PathExists = partial(_path_exists, parser=parser)
    parser.add_argument("pipeline_config_file", type=PathExists, help="Pipeline Config File")
    return parser


def create_transforms(pipeline_config_file):
  
    panpipeconfig_file=str(pipeline_config_file)
    labels_dict =None
    if os.path.exists(pipeline_config_file):
        IFLOGGER.info(f"{pipeline_config_file} exists.")
        with open(pipeline_config_file,'r') as infile:
            labels_dict = json.load(infile)
    
    try:
        
        t1w=getGlob(substitute_labels("<BIDS_DIR>/sub-<PARTICIPANT_LABEL>/ses-<PARTICIPANT_SESSION>/anat/*_T1w.nii.gz",labels_dict))
        # if we cant' find T1w then this must be a situation where we have to use the second session MPRAGE
        if not t1w:
            t1w=getGlob(substitute_labels("<BIDS_DIR>/sub-<PARTICIPANT_LABEL>/*/anat/*_T1w.nii*",labels_dict))
        # use wildcard in QSIANAT
        qsianat=getGlob(getParams(labels_dict,"QSIANAT"))
        t1w_t1acpc=getParams(labels_dict,"T1W_T1ACPC")
        command_base, container = getContainer(labels_dict,nodename="additional_transforms",SPECIFIC="DUMMY_CONTAINER",LOGGER=IFLOGGER)
        transform_complete=ants_registration_rigid(t1w,qsianat,t1w_t1acpc,command_base)


    except Exception as e:
        message = f"Exception caught: \n{e}"
        IFLOGGER.error(message)


def main():
    parser=parse_params()
    args, unknown_args = parser.parse_known_args()
    pipeline_config_file = str(args.pipeline_config_file)

    create_transforms(pipeline_config_file)


# This is the standard boilerplate that calls the main() function.
if __name__ == '__main__':
    main()



from nipype.interfaces.base import BaseInterfaceInputSpec, BaseInterface, File, TraitedSpec, traits
from nipype import Node
from panpipelines.utils.util_functions import *
import os
import glob
from bids import BIDSLayout
import shlex
import subprocess
from nipype import logging as nlogging
import datetime

IFLOGGER=nlogging.getLogger('nipype.interface')

def freesurferextra_proc(labels_dict,subjects_dir=""):

    command_base, container = getContainer(labels_dict,nodename="freesurferextra", SPECIFIC="FREESURFER_CONTAINER",LOGGER=IFLOGGER) 
    participant_label = getParams(labels_dict,'PARTICIPANT_LABEL')
    session_label = getParams(labels_dict,'PARTICIPANT_SESSION')

    subject="sub-"+participant_label
    session="ses-"+session_label

    resultsmessage=[]
    resultsmessage.append(f"Started freesurfer extra processing for {subject},{session}\n")
    resultsmessage.append(f"Datetime started: {str(datetime.datetime.now())}\n")

    cwd=os.getcwd()
    labels_dict = updateParams(labels_dict,"CWD",cwd)
    if not subjects_dir:
        subjects_dir = getParams(labels_dict,'SUBJECTS_DIR')

    os.environ["SUBJECTS_DIR"]=subjects_dir
    os.environ["SINGULARITYENV_SUBJECTS_DIR"]=translate_binding(command_base,subjects_dir)
  
    FREEVER="Unknown"
    IFLOGGER.info("Checking the recon-all version:")
    command = f"{command_base} recon-all --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    if "-7.3.2-" in results:
        FREEVER="7.3.2"
    resultsmessage.append(results)

    IFLOGGER.info("\nChecking the container version:")
    command = f"{command_base} --version"
    evaluated_command=substitute_labels(command, labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    resultsmessage.append(results)

    # create lobe lh
    command=f"{command_base} mri_annotation2label --subject {subject} --hemi lh  --lobesStrict lh.lobes.annot"
    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    resultsmessage.append(results)

    # create lobe rh
    command=f"{command_base} mri_annotation2label --subject {subject} --hemi rh  --lobesStrict rh.lobes.annot"
    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    resultsmessage.append(results)

    lobestats_lh=os.path.join(subjects_dir,subject, "stats","lh.lobe.stats")
    command=f"{command_base} mris_anatomical_stats -a lobes.annot -f {lobestats_lh} {subject} lh"
    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    resultsmessage.append(results)

    lobestats_rh=os.path.join(subjects_dir,subject, "stats","rh.lobe.stats")
    command=f"{command_base} mris_anatomical_stats -a lobes.annot -f {lobestats_rh} {subject} rh"
    evaluated_command=substitute_labels(command,labels_dict)
    results = runCommand(evaluated_command,IFLOGGER)
    resultsmessage.append(results)

    resultsmessage.append(f"Completed freesurfer extra processing for {subject},{session}\n")
    resultsmessage.append(f"\nDatetime Completed: {str(datetime.datetime.now())}\n")

    resultsfile=os.path.join(cwd,f"{subject}_{session}_completed")
    with open(resultsfile,"w") as outfile:
        outfile.write("\n".join(resultsmessage))
    
    out_files=[]
    out_files.insert(0,lobestats_lh)
    out_files.insert(1,lobestats_rh)


    return {
        "output_dir":output_dir,
        "out_files":out_files
    }



class freesurferextraInputSpec(BaseInterfaceInputSpec):
    labels_dict = traits.Dict({},mandatory=False,desc='labels', usedefault=True)
    subjects_dir = traits.String("",desc="Freesurfer Directory", usedefault=True)

class freesurferextraOutputSpec(TraitedSpec):
    output_dir = traits.String(desc="freesurferextra output directory")
    out_files = traits.List(desc='list of files')
    
class freesurferextra_pan(BaseInterface):
    input_spec = freesurferextraInputSpec
    output_spec = freesurferextraOutputSpec

    def _run_interface(self, runtime):

        # Call our python code here:
        outputs = freesurferextra_proc(
            self.inputs.labels_dict,
            self.inputs.bids_dir
        )

        setattr(self, "_results", outputs)
        # And we are done
        return runtime

    def _list_outputs(self):
        return self._results


def create(labels_dict,name='freesurferextra_node',subjects_dir="", LOGGER=IFLOGGER):
    # Create Node
    pan_node = Node(freesurferextra_pan(), name=name)

    if LOGGER:
        LOGGER.info(f"Created Node {pan_node!r}")
        
    # Specify node inputs
    pan_node.inputs.labels_dict = labels_dict
    
    if subjects_dir is None or subjects_dir == "":
        subjects_dir = substitute_labels("<SUBJECTS_DIR>", labels_dict)

    pan_node.inputs.subjects_dir =  subjects_dir

    return pan_node





from urllib.parse import urlparse
from panpipelines.utils.util_functions import *
from panpipelines.utils.upload_functions import *
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from functools import partial
import json
import os
import tempfile
import glob
import multiprocessing as mp
import pandas as pd
import shutil


def parse_params():
    parser = ArgumentParser(description="Pan pipelines")
    PathExists = partial(path_exists, parser=parser)
    parser.add_argument("--ftpcredentials", help="sftp connection string")
    parser.add_argument("--urlstring", help="sftp connection string")
    parser.add_argument("--bids_dir", type=PathExists, help="bids directory")
    parser.add_argument("--participants_file", type=PathExists, help="participants_file in tsv format")
    parser.add_argument("--participant_label", nargs="*", type=drop_sub, help="filter by subject label (the sub- prefix can be removed).")
    parser.add_argument("--all_bids",default="False")
    parser.add_argument("--replace",default="True")
    parser.add_argument("--debug",default="False")
    parser.add_argument("--num_procs",type=int,default=5)
    parser.add_argument("--ftp_raw_path", help="participants_file in tsv format",default="/shared/PAN/PAN-Data/Core-E/Raw")
    parser.add_argument("--ftp_processed_path", help="participants_file in tsv format",default="/shared/PAN/PAN-Data/Core-E/Processed")
    parser.add_argument("--pipeline_config_file", type=Path, help="Pipeline Config File")

    return parser

if __name__ == "__main__":

    parser=parse_params()
    args, unknown_args = parser.parse_known_args()

    raw_path = args.ftp_raw_path
    processed_path = args.ftp_processed_path

    urlstring = args.urlstring
    if urlstring:
        parsed_url = urlparse(urlstring)
        hostname=parsed_url.hostname
        username=parsed_url.username
        password=parsed_url.password
        port=parsed_url.port
        if not port:
            port = 22

    ftpcredentials = args.ftpcredentials
    if ftpcredentials:
        with open(ftpcredentials,"r") as infile:
            ftpcred = json.load(infile)
            hostname=ftpcred["hostname"]
            username=ftpcred["username"]
            password=ftpcred["password"]
            if "port" in ftpcred.keys():
                port=int(ftpcred["password"])
            else:
                port = 22

    pipeline_config_file = None
    if args.pipeline_config_file:
        if Path(args.pipeline_config_file).exists():
            pipeline_config_file = str(args.pipeline_config_file)

    labels_dict={}
    if pipeline_config_file:
        panpipeconfig_file=str(pipeline_config_file)
        if os.path.exists(pipeline_config_file):
           print(f"{pipeline_config_file} exists.")
           with open(pipeline_config_file,'r') as infile:
               labels_dict = json.load(infile)

    participant_label = args.participant_label
    bids_dir = str(args.bids_dir)
    if args.participants_file:
        participants_file = str(args.participants_file)
    else:
        participants_file = None

    ALLBIDS=isTrue(args.all_bids)
    REPLACE=isTrue(args.replace)
    DEBUG=isTrue(args.debug)
    PROCS=args.num_procs

    subject_list = glob.glob(os.path.join(bids_dir,"sub-*"))
    subject_list.sort()
    if ALLBIDS:
        participant_label = None
        participants_file = None
        print("ALLBIDS selected. Ignoring participants labels and participants file")

    folder = os.path.basename(bids_dir)
    if participants_file:
        df = pd.read_table(participants_file,sep="\t")
        participants_file_labels = df["hml_id"].tolist()
        subject_dir_list  = [x for x in subject_list if drop_sub(os.path.basename(x)) in participants_file_labels]
    elif participant_label:
        subject_dir_list  = [x for x in subject_list if drop_sub(os.path.basename(x)) in participant_label]
    else:
        subject_dir_list  = [x for x in subject_list]

    subject_dir_list.sort()
    remote_path_list  = [os.path.join(raw_path,folder,os.path.basename(x)) for x in subject_dir_list]

    multiproc_zip = zip(subject_dir_list,remote_path_list)
    if not DEBUG:
        par_upload_subbids=partial(ftp_upload_subjectbids, hostname=hostname,username=username,password=password,port=port,replace=REPLACE)
        with mp.Pool(PROCS) as pool:
            completedlist = pool.starmap(par_upload_subbids,multiproc_zip)
            pool.close()
            pool.join()
    else:
        for i,(source_path,remote_path) in enumerate(multiproc_zip):
            ftp_upload_subjectbids(source_path,remote_path,hostname=hostname,username=username,password=password,port=port,replace=REPLACE)


    remote_path = os.path.join(raw_path,folder)
    local_participantsTSV = os.path.join(bids_dir,"participants.tsv")
    remote_participantsTSV=os.path.join(remote_path,"participants.tsv")
    ftp_upload(local_participantsTSV ,remote_participantsTSV,hostname,username,password,port)

    local_datasetDescription = os.path.join(bids_dir,"dataset_description.json")
    remote_datasetDescription = os.path.join(remote_path,"dataset_description.json")
    ftp_upload(local_datasetDescription ,remote_datasetDescription,hostname,username,password,port)

    if labels_dict:
        cwd = getParams(labels_dict,"CWD")
    else:
        cwd = os.path.dirname(tempfile.mkstemp()[1])
    metadata_init={}
    remote_metadata_file = newfile(assocfile=os.path.dirname(remote_participantsTSV),suffix="upload-metadata",extension="json")
    local_metadata_file = newfile(outputdir=cwd,assocfile=remote_metadata_file,extension="json")

    history={}
    history["SourceBIDS"]=f"{bids_dir}"
    history["TargetBIDS"]=f"{remote_path}"

    if ALLBIDS:
        history["Description"]=f"All bids files uploaded from {bids_dir}. {len(subject_list)} participants uploaded from {os.path.basename(subject_list[0])} to {os.path.basename(subject_list[-1])}"
    else:
        if participants_file:
            participant_source = participants_file
        else:
            participant_source = participant_label
        history["Description"] =f"Incremental upload from {bids_dir} using {participant_source}. {len(subject_dir_list)} participants uploaded from {os.path.basename(subject_dir_list[0])} to {os.path.basename(subject_dir_list[-1])}"


    upload_metadata(local_metadata_file,remote_metadata_file,local_participantsTSV,remote_participantsTSV,metadata=metadata_init,history=history,hostname=hostname,username=username,password=password, port=port)


    if labels_dict:
        labels_dict["METADATA_FILE"]=local_metadata_file
        labels_dict["OUTPUT_FILE"]=local_participantsTSV
        export_labels(labels_dict,pipeline_config_file)










from panpipelines.utils.util_functions import *
from subprocess import run
import os
import glob


class panpipeline:

    def __init__(self, labels_dict,pipeline_dir, participant_label, name='template_panpipeline', createGraph=True, LOGGER=None, execution={},analysis_level="participant", participant_project=None, participant_session=None,panworkflow = None):
        self.labels_dict = labels_dict
        self.pipeline_dir = pipeline_dir
        self.participant_label = participant_label
        self.name = name
        self.createGraph=createGraph
        self.execution = execution
        self.LOGGER = LOGGER
        self.analysis_level = analysis_level
        self.participant_project = participant_project
        self.participant_session = participant_session
        self.panworkflow = panworkflow
        self.results={}


    def pre_run(self):
        log_message=f"Running {self.name} pipeline at {self.analysis_level} level for "
        if self.analysis_level == "participant":
            log_message = log_message + f"participant {self.participant_label} in "
            if self.participant_session:
                log_message = log_message + f"session {self.participant_session} in "
                if self.participant_project:
                    log_message = log_message + f"project {self.participant_project}\n"
        else:
            participant_info_pairing = [ it for it in zip(self.participant_label,self.participant_project, self.participant_session)]
            for part_info in participant_info_pairing:
                log_message = log_message + f"participant {part_info[0]} in "
                if part_info[1]:
                    log_message = log_message + f"session {part_info[1]} in "
                    if part_info[2]:
                        log_message = log_message + f"project {part_info[2]}\n" 
        if self.LOGGER:
            self.LOGGER.info(log_message)     

    def proc(self):
        workflow_dir = self.pipeline_dir
        workflow_name = "{}_wf".format(self.name)

        pan_workflow = self.panworkflow.create(workflow_name,workflow_dir,self.labels_dict,createGraph=self.createGraph,execution=self.execution, LOGGER=self.LOGGER)
        if isTrue(getParams(self.labels_dict,"FORCE_RUN")):
            MULTIPROC_NPROCS=getParams(self.labels_dict,"MULTIPROC_NPROCS")
            if MULTIPROC_NPROCS:
                MULTIPROC_NPROCS = int(MULTIPROC_NPROCS)
            else:
                MULTIPROC_NPROCS=1
            if self.LOGGER:
                self.LOGGER.info("Forced rerun of workflow selected ....")
            pan_workflow.run(plugin="MultiProc", plugin_args={"overwrite": True,'n_procs' : MULTIPROC_NPROCS})
        else:
            pan_workflow.run()

    def run(self):

        self.pre_run()

        self.proc()

        self.post_run()

        results = self.get_results()
        if self.LOGGER:
            if results:
                self.LOGGER.info(results)
        return results

    def post_run(self):
        log_message=f"Completed {self.name} pipeline at {self.analysis_level} level for "
        if self.analysis_level == "participant":
            log_message = log_message + f"participant {self.participant_label} in "
            if self.participant_session:
                log_message = log_message + f"session {self.participant_session} in "
                if self.participant_project:
                    log_message = log_message + f"project {self.participant_project} "
        else:
            participant_info_pairing = [ it for it in zip(self.participant_label,self.participant_project, self.participant_session)]
            for part_info in participant_info_pairing:
                log_message = log_message + f"participant {part_info[0]} in "
                if part_info[1]:
                    log_message = log_message + f"session {part_info[1]} in "
                    if part_info[2]:
                        log_message = log_message + f"project {part_info[2]} \n" 
        if self.LOGGER:
            self.LOGGER.info(log_message)          


    def get_results(self):
        return self.results





